# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import os
import re
import shutil
import sys
from pathlib import Path
from typing import Any

import pytest

from playwright._impl._path_utils import get_file_dirname
from playwright.async_api import Error, FilePayload, Page
from tests.server import Server
from tests.utils import chromium_version_less_than, must

_dirname = get_file_dirname()
FILE_TO_UPLOAD = _dirname / ".." / "assets/file-to-upload.txt"


async def test_should_upload_the_file(page: Page, server: Server) -> None:
    await page.goto(server.PREFIX + "/input/fileupload.html")
    file_path = os.path.relpath(FILE_TO_UPLOAD, os.getcwd())
    input = await page.query_selector("input")
    assert input
    await input.set_input_files(file_path)
    assert await page.evaluate("e => e.files[0].name", input) == "file-to-upload.txt"
    assert (
        await page.evaluate(
            """e => {
        reader = new FileReader()
        promise = new Promise(fulfill => reader.onload = fulfill)
        reader.readAsText(e.files[0])
        return promise.then(() => reader.result)
    }""",
            input,
        )
        == "contents of the file\n"
    )


async def test_should_work(page: Page, assetdir: Path) -> None:
    await page.set_content("<input type=file>")
    await page.set_input_files("input", assetdir / "file-to-upload.txt")
    assert await page.eval_on_selector("input", "input => input.files.length") == 1
    assert (
        await page.eval_on_selector("input", "input => input.files[0].name")
        == "file-to-upload.txt"
    )


async def test_should_set_from_memory(page: Page) -> None:
    await page.set_content("<input type=file>")
    file: FilePayload = {
        "name": "test.txt",
        "mimeType": "text/plain",
        "buffer": b"this is a test",
    }
    await page.set_input_files(
        "input",
        files=[file],
    )
    assert await page.eval_on_selector("input", "input => input.files.length") == 1
    assert (
        await page.eval_on_selector("input", "input => input.files[0].name")
        == "test.txt"
    )


async def test_should_emit_event(page: Page) -> None:
    await page.set_content("<input type=file>")
    fc_done: asyncio.Future = asyncio.Future()
    page.once("filechooser", lambda file_chooser: fc_done.set_result(file_chooser))
    await page.click("input")
    file_chooser = await fc_done
    assert file_chooser
    assert (
        repr(file_chooser)
        == f"<FileChooser page={file_chooser.page} element={file_chooser.element}>"
    )


async def test_should_work_when_file_input_is_attached_to_dom(page: Page) -> None:
    await page.set_content("<input type=file>")
    async with page.expect_file_chooser() as fc_info:
        await page.click("input")
    file_chooser = await fc_info.value
    assert file_chooser


async def test_should_work_when_file_input_is_not_attached_to_DOM(page: Page) -> None:
    async with page.expect_file_chooser() as fc_info:
        await page.evaluate(
            """() => {
                el = document.createElement('input')
                el.type = 'file'
                el.click()
            }"""
        )
    file_chooser = await fc_info.value
    assert file_chooser


async def test_should_return_the_same_file_chooser_when_there_are_many_watchdogs_simultaneously(
    page: Page,
) -> None:
    await page.set_content("<input type=file>")
    results = await asyncio.gather(
        page.wait_for_event("filechooser"),
        page.wait_for_event("filechooser"),
        page.eval_on_selector("input", "input => input.click()"),
    )
    assert results[0] == results[1]


async def test_should_accept_single_file(page: Page) -> None:
    await page.set_content('<input type=file oninput="javascript:console.timeStamp()">')
    async with page.expect_file_chooser() as fc_info:
        await page.click("input")
    file_chooser = await fc_info.value
    assert file_chooser.page == page
    assert file_chooser.element
    await file_chooser.set_files(FILE_TO_UPLOAD)
    assert await page.eval_on_selector("input", "input => input.files.length") == 1
    assert (
        await page.eval_on_selector("input", "input => input.files[0].name")
        == "file-to-upload.txt"
    )


async def test_should_be_able_to_read_selected_file(page: Page) -> None:
    page.once(
        "filechooser", lambda file_chooser: file_chooser.set_files(FILE_TO_UPLOAD)
    )
    await page.set_content("<input type=file>")
    content = await page.eval_on_selector(
        "input",
        """async picker => {
            picker.click();
            await new Promise(x => picker.oninput = x);
            const reader = new FileReader();
            const promise = new Promise(fulfill => reader.onload = fulfill);
            reader.readAsText(picker.files[0]);
            return promise.then(() => reader.result);
        }""",
    )
    assert content == "contents of the file\n"


async def test_should_be_able_to_reset_selected_files_with_empty_file_list(
    page: Page,
) -> None:
    await page.set_content("<input type=file>")
    page.once(
        "filechooser", lambda file_chooser: file_chooser.set_files(FILE_TO_UPLOAD)
    )
    file_length = 0
    async with page.expect_file_chooser():
        file_length = await page.eval_on_selector(
            "input",
            """async picker => {
                picker.click();
                await new Promise(x => picker.oninput = x);
                return picker.files.length;
            }""",
        )
    assert file_length == 1

    page.once("filechooser", lambda file_chooser: file_chooser.set_files([]))
    async with page.expect_file_chooser():
        file_length = await page.eval_on_selector(
            "input",
            """async picker => {
                picker.click();
                await new Promise(x => picker.oninput = x);
                return picker.files.length;
            }""",
        )
    assert file_length == 0


async def test_should_not_accept_multiple_files_for_single_file_input(
    page: Page, assetdir: Path
) -> None:
    await page.set_content("<input type=file>")
    async with page.expect_file_chooser() as fc_info:
        await page.click("input")
    file_chooser = await fc_info.value
    with pytest.raises(Exception) as exc_info:
        await file_chooser.set_files(
            [
                os.path.realpath(assetdir / "file-to-upload.txt"),
                os.path.realpath(assetdir / "pptr.png"),
            ]
        )
    assert exc_info.value


async def test_should_emit_input_and_change_events(page: Page) -> None:
    events = []
    await page.expose_function("eventHandled", lambda e: events.append(e))
    await page.set_content(
        """
            <input id=input type=file></input>
            <script>
            input.addEventListener('input', e => eventHandled({ type: e.type }))
            input.addEventListener('change', e => eventHandled({ type: e.type }))
            </script>
        """
    )
    await must(await page.query_selector("input")).set_input_files(FILE_TO_UPLOAD)
    assert len(events) == 2
    assert events[0]["type"] == "input"
    assert events[1]["type"] == "change"


async def test_should_work_for_single_file_pick(page: Page) -> None:
    await page.set_content("<input type=file>")
    async with page.expect_file_chooser() as fc_info:
        await page.click("input")
    file_chooser = await fc_info.value
    assert file_chooser.is_multiple() is False


async def test_should_work_for_multiple(page: Page) -> None:
    await page.set_content("<input multiple type=file>")
    async with page.expect_file_chooser() as fc_info:
        await page.click("input")
    file_chooser = await fc_info.value
    assert file_chooser.is_multiple()


async def test_should_work_for_webkitdirectory(page: Page) -> None:
    await page.set_content("<input multiple webkitdirectory type=file>")
    async with page.expect_file_chooser() as fc_info:
        await page.click("input")
    file_chooser = await fc_info.value
    assert file_chooser.is_multiple()


def _assert_wheel_event(expected: Any, received: Any, browser_name: str) -> None:
    # Chromium reports deltaX/deltaY scaled by host device scale factor.
    # https://bugs.chromium.org/p/chromium/issues/detail?id=1324819
    # https://github.com/microsoft/playwright/issues/7362
    # Different bots have different scale factors (usually 1 or 2), so we just ignore the values
    # instead of guessing the host scale factor.
    if sys.platform == "darwin" and browser_name == "chromium":
        del expected["deltaX"]
        del expected["deltaY"]
        del received["deltaX"]
        del received["deltaY"]
    assert received == expected


async def test_wheel_should_work(page: Page, browser_name: str) -> None:
    await page.set_content(
        """
        <div style="width: 5000px; height: 5000px;"></div>
    """
    )
    await page.mouse.move(50, 60)
    await _listen_for_wheel_events(page, "div")
    await page.mouse.wheel(0, 100)
    _assert_wheel_event(
        await page.evaluate("window.lastEvent"),
        {
            "deltaX": 0,
            "deltaY": 100,
            "clientX": 50,
            "clientY": 60,
            "deltaMode": 0,
            "ctrlKey": False,
            "shiftKey": False,
            "altKey": False,
            "metaKey": False,
        },
        browser_name,
    )
    await page.wait_for_function("window.scrollY === 100")


async def _listen_for_wheel_events(page: Page, selector: str) -> None:
    await page.evaluate(
        """
        selector => {
            document.querySelector(selector).addEventListener('wheel', (e) => {
            window['lastEvent'] = {
                deltaX: e.deltaX,
                deltaY: e.deltaY,
                clientX: e.clientX,
                clientY: e.clientY,
                deltaMode: e.deltaMode,
                ctrlKey: e.ctrlKey,
                shiftKey: e.shiftKey,
                altKey: e.altKey,
                metaKey: e.metaKey,
            };
            }, { passive: false });
        }
    """,
        selector,
    )


async def test_should_upload_large_file(
    page: Page, server: Server, tmp_path: Path
) -> None:
    await page.goto(server.PREFIX + "/input/fileupload.html")
    large_file_path = tmp_path / "200MB.zip"
    data = b"A" * 1024
    with large_file_path.open("wb") as f:
        for i in range(0, 200 * 1024 * 1024, len(data)):
            f.write(data)
    input = page.locator('input[type="file"]')
    events = await input.evaluate_handle(
        """
        e => {
            const events = [];
            e.addEventListener('input', () => events.push('input'));
            e.addEventListener('change', () => events.push('change'));
            return events;
        }
    """
    )

    await input.set_input_files(large_file_path)
    assert await input.evaluate("e => e.files[0].name") == "200MB.zip"
    assert await events.evaluate("e => e") == ["input", "change"]

    [request, _] = await asyncio.gather(
        server.wait_for_request("/upload"),
        page.click("input[type=submit]"),
    )

    contents = request.args[b"file1"][0]
    assert len(contents) == 200 * 1024 * 1024
    assert contents[:1024] == data
    # flake8: noqa: E203
    assert contents[len(contents) - 1024 :] == data
    assert request.post_body
    match = re.search(
        rb'^.*Content-Disposition: form-data; name="(?P<name>.*)"; filename="(?P<filename>.*)".*$',
        request.post_body,
        re.MULTILINE,
    )
    assert match
    assert match.group("name") == b"file1"
    assert match.group("filename") == b"200MB.zip"


async def test_set_input_files_should_preserve_last_modified_timestamp(
    page: Page,
    assetdir: Path,
) -> None:
    await page.set_content("<input type=file multiple=true/>")
    input = page.locator("input")
    files = ["file-to-upload.txt", "file-to-upload-2.txt"]
    await input.set_input_files([assetdir / file for file in files])
    assert await input.evaluate("input => [...input.files].map(f => f.name)") == files
    timestamps = await input.evaluate(
        "input => [...input.files].map(f => f.lastModified)"
    )
    expected_timestamps = [os.path.getmtime(assetdir / file) * 1000 for file in files]

    # On Linux browser sometimes reduces the timestamp by 1ms: 1696272058110.0715  -> 1696272058109 or even
    # rounds it to seconds in WebKit: 1696272058110 -> 1696272058000.
    for i in range(len(timestamps)):
        assert abs(timestamps[i] - expected_timestamps[i]) < 1000


async def test_should_upload_multiple_large_file(
    page: Page, server: Server, tmp_path: Path
) -> None:
    files_count = 10
    await page.goto(server.PREFIX + "/input/fileupload-multi.html")
    upload_file = tmp_path / "50MB_1.zip"
    data = b"A" * 1024
    with upload_file.open("wb") as f:
        # 49 is close to the actual limit
        for i in range(0, 49 * 1024):
            f.write(data)
    input = page.locator('input[type="file"]')
    upload_files = [upload_file]
    for i in range(2, files_count + 1):
        dst_file = tmp_path / f"50MB_{i}.zip"
        shutil.copy(upload_file, dst_file)
        upload_files.append(dst_file)
    async with page.expect_file_chooser() as fc_info:
        await input.click()
    file_chooser = await fc_info.value
    await file_chooser.set_files(upload_files)
    files_len = await page.evaluate(
        'document.getElementsByTagName("input")[0].files.length'
    )
    assert file_chooser.is_multiple()
    assert files_len == files_count
    for path in upload_files:
        path.unlink()


async def test_should_upload_a_folder(
    page: Page,
    server: Server,
    tmp_path: Path,
    browser_name: str,
    browser_version: str,
    headless: bool,
) -> None:
    await page.goto(server.PREFIX + "/input/folderupload.html")
    input = await page.query_selector("input")
    assert input
    dir = tmp_path / "file-upload-test"
    dir.mkdir()
    (dir / "file1.txt").write_text("file1 content")
    (dir / "file2").write_text("file2 content")
    (dir / "sub-dir").mkdir()
    (dir / "sub-dir" / "really.txt").write_text("sub-dir file content")
    await input.set_input_files(dir)
    assert set(
        await input.evaluate("e => [...e.files].map(f => f.webkitRelativePath)")
    ) == set(
        [
            "file-upload-test/file1.txt",
            "file-upload-test/file2",
            # https://issues.chromium.org/issues/345393164
            *(
                []
                if browser_name == "chromium"
                and headless
                and chromium_version_less_than(browser_version, "127.0.6533.0")
                else ["file-upload-test/sub-dir/really.txt"]
            ),
        ]
    )
    webkit_relative_paths = await input.evaluate(
        "e => [...e.files].map(f => f.webkitRelativePath)"
    )
    for i, webkit_relative_path in enumerate(webkit_relative_paths):
        content = await input.evaluate(
            """(e, i) => {
            const reader = new FileReader();
            const promise = new Promise(fulfill => reader.onload = fulfill);
            reader.readAsText(e.files[i]);
            return promise.then(() => reader.result);
        }""",
            i,
        )
        assert content == (dir / ".." / webkit_relative_path).read_text()


async def test_should_upload_a_folder_and_throw_for_multiple_directories(
    page: Page, server: Server, tmp_path: Path
) -> None:
    await page.goto(server.PREFIX + "/input/folderupload.html")
    input = page.locator("input")
    dir = tmp_path / "file-upload-test"
    dir.mkdir()
    (dir / "folder1").mkdir()
    (dir / "folder1" / "file1.txt").write_text("file1 content")
    (dir / "folder2").mkdir()
    (dir / "folder2" / "file2.txt").write_text("file2 content")
    with pytest.raises(Error) as exc_info:
        await input.set_input_files([dir / "folder1", dir / "folder2"])
    assert "Multiple directories are not supported" in exc_info.value.message


async def test_should_throw_if_a_directory_and_files_are_passed(
    page: Page, server: Server, tmp_path: Path
) -> None:
    await page.goto(server.PREFIX + "/input/folderupload.html")
    input = page.locator("input")
    dir = tmp_path / "file-upload-test"
    dir.mkdir()
    (dir / "file1.txt").write_text("file1 content")
    with pytest.raises(Error) as exc_info:
        await input.set_input_files([dir, dir / "file1.txt"])
    assert (
        "File paths must be all files or a single directory" in exc_info.value.message
    )


async def test_should_throw_when_upload_a_folder_in_a_normal_file_upload_input(
    page: Page, server: Server, tmp_path: Path
) -> None:
    await page.goto(server.PREFIX + "/input/fileupload.html")
    input = await page.query_selector("input")
    assert input
    dir = tmp_path / "file-upload-test"
    dir.mkdir()
    (dir / "file1.txt").write_text("file1 content")
    with pytest.raises(Error) as exc_info:
        await input.set_input_files(dir)
    assert (
        "File input does not support directories, pass individual files instead"
        in exc_info.value.message
    )



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import pytest

from playwright.async_api import Browser, BrowserType, Error


async def test_should_create_new_page(browser: Browser) -> None:
    page1 = await browser.new_page()
    assert len(browser.contexts) == 1

    page2 = await browser.new_page()
    assert len(browser.contexts) == 2

    await page1.close()
    assert len(browser.contexts) == 1

    await page2.close()
    assert len(browser.contexts) == 0


async def test_should_throw_upon_second_create_new_page(browser: Browser) -> None:
    page = await browser.new_page()
    with pytest.raises(Error) as exc:
        await page.context.new_page()
    await page.close()
    assert "Please use browser.new_context()" in exc.value.message


async def test_version_should_work(browser: Browser, is_chromium: bool) -> None:
    version = browser.version
    if is_chromium:
        assert re.match(r"^\d+\.\d+\.\d+\.\d+$", version)
    else:
        assert re.match(r"^\d+\.\d+", version)


async def test_should_return_browser_type(
    browser: Browser, browser_type: BrowserType
) -> None:
    assert browser.browser_type is browser_type



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path
from typing import Dict

import pytest

from playwright.sync_api import Browser, BrowserType, Error
from tests.server import Server


def test_should_expose_video_path(
    browser: Browser, tmp_path: Path, server: Server
) -> None:
    page = browser.new_page(
        record_video_dir=tmp_path, record_video_size={"width": 100, "height": 200}
    )
    page.goto(server.PREFIX + "/grid.html")
    video = page.video
    assert video
    path = video.path()
    assert repr(page.video) == f"<Video page={page}>"
    assert str(tmp_path) in str(path)
    page.wait_for_timeout(1000)
    page.context.close()


def test_video_should_exist(browser: Browser, tmp_path: Path, server: Server) -> None:
    page = browser.new_page(record_video_dir=tmp_path)
    page.goto(server.PREFIX + "/grid.html")
    video = page.video
    assert video
    path = video.path()
    assert str(tmp_path) in str(path)
    page.wait_for_timeout(1000)
    page.context.close()
    assert os.path.exists(path)


def test_record_video_to_path(browser: Browser, tmp_path: Path, server: Server) -> None:
    page = browser.new_page(record_video_dir=tmp_path)
    page.goto(server.PREFIX + "/grid.html")
    video = page.video
    assert video
    path = video.path()
    assert str(tmp_path) in str(path)
    page.wait_for_timeout(1000)
    page.context.close()
    assert os.path.exists(path)


def test_record_video_to_path_persistent(
    browser_type: BrowserType, tmp_path: Path, server: Server, launch_arguments: Dict
) -> None:
    context = browser_type.launch_persistent_context(
        tmp_path, **launch_arguments, record_video_dir=tmp_path
    )
    page = context.pages[0]
    page.goto(server.PREFIX + "/grid.html")
    video = page.video
    assert video
    path = video.path()
    assert str(tmp_path) in str(path)
    page.wait_for_timeout(1000)
    context.close()
    assert os.path.exists(path)


def test_record_video_can_get_video_path_immediately(
    browser_type: BrowserType, tmp_path: Path, launch_arguments: Dict
) -> None:
    context = browser_type.launch_persistent_context(
        tmp_path, **launch_arguments, record_video_dir=tmp_path
    )
    page = context.pages[0]
    video = page.video
    assert video
    path = video.path()
    assert str(tmp_path) in str(path)
    page.wait_for_timeout(1000)
    context.close()
    assert os.path.exists(path)


def test_should_error_if_page_not_closed_before_save_as(
    browser: Browser, tmp_path: Path, server: Server
) -> None:
    page = browser.new_page(record_video_dir=tmp_path)
    page.goto(server.PREFIX + "/grid.html")
    page.wait_for_timeout(1000)  # Give it some time to record.
    out_path = tmp_path / "some-video.webm"
    with pytest.raises(Error) as err:
        video = page.video
        assert video
        video.save_as(out_path)
    assert video
    assert "Page is not yet closed. Close the page prior to calling save_as" in str(err)
    assert not os.path.exists(out_path)
    page.close()

    video.save_as(out_path)
    assert os.path.exists(out_path)



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from typing import List

import pytest

from playwright.async_api import ElementHandle, Error, Page
from tests.server import Server


async def give_it_a_chance_to_resolve(page: Page) -> None:
    for i in range(5):
        await page.evaluate(
            "() => new Promise(f => requestAnimationFrame(() => requestAnimationFrame(f)))"
        )


async def wait_for_state(div: ElementHandle, state: str, done: List[bool]) -> None:
    await div.wait_for_element_state(state)  # type: ignore
    done[0] = True


async def wait_for_state_to_throw(
    div: ElementHandle, state: str
) -> pytest.ExceptionInfo[Error]:
    with pytest.raises(Error) as exc_info:
        await div.wait_for_element_state(state)  # type: ignore
    return exc_info


async def test_should_wait_for_visible(page: Page) -> None:
    await page.set_content('<div style="display:none">content</div>')
    div = await page.query_selector("div")
    assert div
    done = [False]
    promise = asyncio.create_task(wait_for_state(div, "visible", done))
    await give_it_a_chance_to_resolve(page)
    assert done[0] is False
    assert div
    await div.evaluate('div => div.style.display = "block"')
    await promise


async def test_should_wait_for_already_visible(page: Page) -> None:
    await page.set_content("<div>content</div>")
    div = await page.query_selector("div")
    assert div
    await div.wait_for_element_state("visible")


async def test_should_timeout_waiting_for_visible(page: Page) -> None:
    await page.set_content('<div style="display:none">content</div>')
    div = await page.query_selector("div")
    assert div
    with pytest.raises(Error) as exc_info:
        await div.wait_for_element_state("visible", timeout=1000)
    assert "Timeout 1000ms exceeded" in exc_info.value.message


async def test_should_throw_waiting_for_visible_when_detached(page: Page) -> None:
    await page.set_content('<div style="display:none">content</div>')
    div = await page.query_selector("div")
    assert div
    promise = asyncio.create_task(wait_for_state_to_throw(div, "visible"))
    await div.evaluate("div => div.remove()")
    exc_info = await promise
    assert "Element is not attached to the DOM" in exc_info.value.message


async def test_should_wait_for_hidden(page: Page) -> None:
    await page.set_content("<div>content</div>")
    div = await page.query_selector("div")
    assert div
    done = [False]
    promise = asyncio.create_task(wait_for_state(div, "hidden", done))
    await give_it_a_chance_to_resolve(page)
    assert done[0] is False
    await div.evaluate('div => div.style.display = "none"')
    await promise


async def test_should_wait_for_already_hidden(page: Page) -> None:
    await page.set_content("<div></div>")
    div = await page.query_selector("div")
    assert div
    await div.wait_for_element_state("hidden")


async def test_should_wait_for_hidden_when_detached(page: Page) -> None:
    await page.set_content("<div>content</div>")
    div = await page.query_selector("div")
    assert div
    done = [False]
    promise = asyncio.create_task(wait_for_state(div, "hidden", done))
    await give_it_a_chance_to_resolve(page)
    assert done[0] is False
    assert div
    await div.evaluate("div => div.remove()")
    await promise


async def test_should_wait_for_enabled_button(page: Page, server: Server) -> None:
    await page.set_content("<button disabled><span>Target</span></button>")
    span = await page.query_selector("text=Target")
    assert span
    done = [False]
    promise = asyncio.create_task(wait_for_state(span, "enabled", done))
    await give_it_a_chance_to_resolve(page)
    assert done[0] is False
    await span.evaluate("span => span.parentElement.disabled = false")
    await promise


async def test_should_throw_waiting_for_enabled_when_detached(page: Page) -> None:
    await page.set_content("<button disabled>Target</button>")
    button = await page.query_selector("button")
    assert button
    promise = asyncio.create_task(wait_for_state_to_throw(button, "enabled"))
    await button.evaluate("button => button.remove()")
    exc_info = await promise
    assert "Element is not attached to the DOM" in exc_info.value.message


async def test_should_wait_for_disabled_button(page: Page) -> None:
    await page.set_content("<button><span>Target</span></button>")
    span = await page.query_selector("text=Target")
    assert span
    done = [False]
    promise = asyncio.create_task(wait_for_state(span, "disabled", done))
    await give_it_a_chance_to_resolve(page)
    assert done[0] is False
    await span.evaluate("span => span.parentElement.disabled = true")
    await promise


async def test_should_wait_for_editable_input(page: Page, server: Server) -> None:
    await page.set_content("<input readonly>")
    input = await page.query_selector("input")
    assert input
    done = [False]
    promise = asyncio.create_task(wait_for_state(input, "editable", done))
    await give_it_a_chance_to_resolve(page)
    assert done[0] is False
    await input.evaluate("input => input.readOnly = false")
    await promise



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Optional

from playwright.async_api import BrowserContext, Error, Page


async def test_add_init_script_evaluate_before_anything_else_on_the_page(
    page: Page,
) -> None:
    await page.add_init_script("window.injected = 123")
    await page.goto("data:text/html,<script>window.result = window.injected</script>")
    assert await page.evaluate("window.result") == 123


async def test_add_init_script_work_with_a_path(page: Page, assetdir: Path) -> None:
    await page.add_init_script(path=assetdir / "injectedfile.js")
    await page.goto("data:text/html,<script>window.result = window.injected</script>")
    assert await page.evaluate("window.result") == 123


async def test_add_init_script_work_with_content(page: Page) -> None:
    await page.add_init_script("window.injected = 123")
    await page.goto("data:text/html,<script>window.result = window.injected</script>")
    assert await page.evaluate("window.result") == 123


async def test_add_init_script_throw_without_path_and_content(page: Page) -> None:
    error: Optional[Error] = None
    try:
        await page.add_init_script({"foo": "bar"})  # type: ignore
    except Error as e:
        error = e
    assert error
    assert error.message == "Either path or script parameter must be specified"


async def test_add_init_script_work_with_browser_context_scripts(
    page: Page, context: BrowserContext
) -> None:
    await context.add_init_script("window.temp = 123")
    page = await context.new_page()
    await page.add_init_script("window.injected = window.temp")
    await page.goto("data:text/html,<script>window.result = window.injected</script>")
    assert await page.evaluate("window.result") == 123


async def test_add_init_script_work_with_browser_context_scripts_with_a_path(
    page: Page, context: BrowserContext, assetdir: Path
) -> None:
    await context.add_init_script(path=assetdir / "injectedfile.js")
    page = await context.new_page()
    await page.goto("data:text/html,<script>window.result = window.injected</script>")
    assert await page.evaluate("window.result") == 123


async def test_add_init_script_work_with_browser_context_scripts_for_already_created_pages(
    page: Page, context: BrowserContext
) -> None:
    await context.add_init_script("window.temp = 123")
    await page.add_init_script("window.injected = window.temp")
    await page.goto("data:text/html,<script>window.result = window.injected</script>")
    assert await page.evaluate("window.result") == 123


async def test_add_init_script_support_multiple_scripts(page: Page) -> None:
    await page.add_init_script("window.script1 = 1")
    await page.add_init_script("window.script2 = 2")
    await page.goto("data:text/html,<script>window.result = window.injected</script>")
    assert await page.evaluate("window.script1") == 1
    assert await page.evaluate("window.script2") == 2


async def test_should_work_with_trailing_comments(page: Page) -> None:
    await page.add_init_script("// comment")
    await page.add_init_script("window.secret = 42;")
    await page.goto("data:text/html,<html></html>")
    assert await page.evaluate("secret") == 42



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, List

import pytest

from playwright.sync_api import BrowserContext, Error, Page, Request, Route
from tests.server import Server


def _append_with_return_value(values: List, value: Any) -> Any:
    values.append(value)


def test_should_work(page: Page, context: BrowserContext, server: Server) -> None:
    context.route("**/*", lambda route: route.fallback())
    page.goto(server.EMPTY_PAGE)


def test_should_fall_back(page: Page, context: BrowserContext, server: Server) -> None:
    intercepted: List[int] = []
    context.route(
        "**/empty.html",
        lambda route: (
            _append_with_return_value(intercepted, 1),
            route.fallback(),
        ),
    )
    context.route(
        "**/empty.html",
        lambda route: (
            _append_with_return_value(intercepted, 2),
            route.fallback(),
        ),
    )
    context.route(
        "**/empty.html",
        lambda route: (
            _append_with_return_value(intercepted, 3),
            route.fallback(),
        ),
    )

    page.goto(server.EMPTY_PAGE)
    assert intercepted == [3, 2, 1]


def test_should_fall_back_async_delayed(
    page: Page, context: BrowserContext, server: Server
) -> None:
    intercepted = []

    def create_handler(i: int) -> Callable[[Route], None]:
        def handler(route: Route) -> None:
            intercepted.append(i)
            page.wait_for_timeout(500)
            route.fallback()

        return handler

    context.route("**/empty.html", create_handler(1))
    context.route("**/empty.html", create_handler(2))
    context.route("**/empty.html", create_handler(3))
    page.goto(server.EMPTY_PAGE)
    assert intercepted == [3, 2, 1]


def test_should_chain_once(page: Page, context: BrowserContext, server: Server) -> None:
    context.route(
        "**/madeup.txt",
        lambda route: route.fulfill(status=200, body="fulfilled one"),
        times=1,
    )
    context.route("**/madeup.txt", lambda route: route.fallback(), times=1)

    resp = page.goto(server.PREFIX + "/madeup.txt")
    assert resp
    body = resp.body()
    assert body == b"fulfilled one"


def test_should_not_chain_fulfill(
    page: Page, context: BrowserContext, server: Server
) -> None:
    failed = [False]

    def handler(route: Route) -> None:
        failed[0] = True

    context.route("**/empty.html", handler)
    context.route(
        "**/empty.html",
        lambda route: route.fulfill(status=200, body="fulfilled"),
    )
    context.route("**/empty.html", lambda route: route.fallback())

    response = page.goto(server.EMPTY_PAGE)
    assert response
    body = response.body()
    assert body == b"fulfilled"
    assert not failed[0]


def test_should_not_chain_abort(
    page: Page,
    context: BrowserContext,
    server: Server,
    is_webkit: bool,
    is_firefox: bool,
) -> None:
    failed = [False]

    def handler(route: Route) -> None:
        failed[0] = True

    context.route("**/empty.html", handler)
    context.route("**/empty.html", lambda route: route.abort())
    context.route("**/empty.html", lambda route: route.fallback())

    with pytest.raises(Error) as excinfo:
        page.goto(server.EMPTY_PAGE)
    if is_webkit:
        assert "Blocked by Web Inspector" in excinfo.value.message
    elif is_firefox:
        assert "NS_ERROR_FAILURE" in excinfo.value.message
    else:
        assert "net::ERR_FAILED" in excinfo.value.message
    assert not failed[0]


def test_should_fall_back_after_exception(
    page: Page, context: BrowserContext, server: Server
) -> None:
    context.route("**/empty.html", lambda route: route.continue_())

    def handler(route: Route) -> None:
        try:
            route.fulfill(response=47)  # type: ignore
        except Exception:
            route.fallback()

    context.route("**/empty.html", handler)

    page.goto(server.EMPTY_PAGE)


def test_should_amend_http_headers(
    page: Page, context: BrowserContext, server: Server
) -> None:
    values = []

    def handler(route: Route) -> None:
        values.append(route.request.headers.get("foo"))
        values.append(route.request.header_value("FOO"))
        route.continue_()

    context.route("**/sleep.zzz", handler)

    def handler_with_header_mods(route: Route) -> None:
        route.fallback(headers={**route.request.headers, "FOO": "bar"})

    context.route("**/*", handler_with_header_mods)

    page.goto(server.EMPTY_PAGE)
    with server.expect_request("/sleep.zzz") as server_request_info:
        page.evaluate("() => fetch('/sleep.zzz')")
    values.append(server_request_info.value.getHeader("foo"))
    assert values == ["bar", "bar", "bar"]


def test_should_delete_header_with_undefined_value(
    page: Page, context: BrowserContext, server: Server
) -> None:
    page.goto(server.EMPTY_PAGE)
    server.set_route(
        "/something",
        lambda r: (
            r.setHeader("Acces-Control-Allow-Origin", "*"),
            r.write(b"done"),
            r.finish(),
        ),
    )

    intercepted_request = []

    def capture_and_continue(route: Route, request: Request) -> None:
        intercepted_request.append(request)
        route.continue_()

    context.route("**/*", capture_and_continue)

    def delete_foo_header(route: Route, request: Request) -> None:
        headers = request.all_headers()
        del headers["foo"]
        route.fallback(headers=headers)

    context.route(server.PREFIX + "/something", delete_foo_header)
    with server.expect_request("/something") as server_req_info:
        text = page.evaluate(
            """
            async url => {
                const data = await fetch(url, {
                    headers: {
                    foo: 'a',
                    bar: 'b',
                    }
                });
                return data.text();
                }
            """,
            server.PREFIX + "/something",
        )
    server_req = server_req_info.value
    assert text == "done"
    assert not intercepted_request[0].headers.get("foo")
    assert intercepted_request[0].headers.get("bar") == "b"
    assert not server_req.getHeader("foo")
    assert server_req.getHeader("bar") == "b"


def test_should_amend_method(
    page: Page, context: BrowserContext, server: Server
) -> None:
    page.goto(server.EMPTY_PAGE)
    method: List[str] = []
    context.route(
        "**/*",
        lambda route: (
            _append_with_return_value(method, route.request.method),
            route.continue_(),
        ),
    )
    context.route("**/*", lambda route: route.fallback(method="POST"))

    with server.expect_request("/sleep.zzz") as request_info:
        page.evaluate("() => fetch('/sleep.zzz')")
    request = request_info.value
    assert method == ["POST"]
    assert request.method == b"POST"


def test_should_override_request_url(
    page: Page, context: BrowserContext, server: Server
) -> None:
    url: List[str] = []
    context.route(
        "**/global-var.html",
        lambda route: (
            _append_with_return_value(url, route.request.url),
            route.continue_(),
        ),
    )
    context.route(
        "**/foo",
        lambda route: route.fallback(url=server.PREFIX + "/global-var.html"),
    )

    with server.expect_request("/global-var.html") as server_request_info:
        with page.expect_event("response") as response_info:
            page.goto(server.PREFIX + "/foo")
    server_request = server_request_info.value
    response = response_info.value
    assert url == [server.PREFIX + "/global-var.html"]
    assert response.url == server.PREFIX + "/global-var.html"
    assert response.request.url == server.PREFIX + "/global-var.html"
    assert page.evaluate("() => window['globalVar']") == 123
    assert server_request.uri == b"/global-var.html"
    assert server_request.method == b"GET"


def test_should_amend_post_data(
    page: Page, context: BrowserContext, server: Server
) -> None:
    page.goto(server.EMPTY_PAGE)
    post_data: List[str] = []
    context.route(
        "**/*",
        lambda route: (
            _append_with_return_value(post_data, route.request.post_data),
            route.continue_(),
        ),
    )
    context.route("**/*", lambda route: route.fallback(post_data="doggo"))

    with server.expect_request("/sleep.zzz") as server_request_info:
        page.evaluate("() => fetch('/sleep.zzz', { method: 'POST', body: 'birdy' })")
    server_request = server_request_info.value
    assert post_data == ["doggo"]
    assert server_request.post_body == b"doggo"


def test_should_amend_binary_post_data(
    page: Page, context: BrowserContext, server: Server
) -> None:
    page.goto(server.EMPTY_PAGE)
    post_data_buffer: List[str] = []
    context.route(
        "**/*",
        lambda route: (
            _append_with_return_value(post_data_buffer, route.request.post_data),
            route.continue_(),
        ),
    )
    context.route(
        "**/*", lambda route: route.fallback(post_data=b"\x00\x01\x02\x03\x04")
    )

    with server.expect_request("/sleep.zzz") as server_request_info:
        page.evaluate("() => fetch('/sleep.zzz', { method: 'POST', body: 'birdy' })")
    server_request = server_request_info.value
    # FIXME: should this be bytes?
    assert post_data_buffer == ["\x00\x01\x02\x03\x04"]
    assert server_request.method == b"POST"
    assert server_request.post_body == b"\x00\x01\x02\x03\x04"


def test_should_chain_fallback_into_page(
    context: BrowserContext, page: Page, server: Server
) -> None:
    intercepted: List[int] = []
    context.route(
        "**/empty.html",
        lambda route: (_append_with_return_value(intercepted, 1), route.fallback()),
    )
    context.route(
        "**/empty.html",
        lambda route: (_append_with_return_value(intercepted, 2), route.fallback()),
    )
    context.route(
        "**/empty.html",
        lambda route: (_append_with_return_value(intercepted, 3), route.fallback()),
    )
    page.route(
        "**/empty.html",
        lambda route: (_append_with_return_value(intercepted, 4), route.fallback()),
    )
    page.route(
        "**/empty.html",
        lambda route: (_append_with_return_value(intercepted, 5), route.fallback()),
    )
    page.route(
        "**/empty.html",
        lambda route: (_append_with_return_value(intercepted, 6), route.fallback()),
    )

    page.goto(server.EMPTY_PAGE)
    assert intercepted == [6, 5, 4, 3, 2, 1]



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import pytest

from playwright.async_api import Page


async def test_accessibility_should_work(
    page: Page, is_firefox: bool, is_chromium: bool, is_webkit: bool
) -> None:
    if is_webkit and sys.platform == "darwin":
        pytest.skip("Test disabled on WebKit on macOS")
    await page.set_content(
        """<head>
      <title>Accessibility Test</title>
    </head>
    <body>
      <h1>Inputs</h1>
      <input placeholder="Empty input" autofocus />
      <input placeholder="readonly input" readonly />
      <input placeholder="disabled input" disabled />
      <input aria-label="Input with whitespace" value="  " />
      <input value="value only" />
      <input aria-placeholder="placeholder" value="and a value" />
      <div aria-hidden="true" id="desc">This is a description!</div>
      <input aria-placeholder="placeholder" value="and a value" aria-describedby="desc" />
    </body>"""
    )
    # autofocus happens after a delay in chrome these days
    await page.wait_for_function("document.activeElement.hasAttribute('autofocus')")

    if is_firefox:
        golden = {
            "role": "document",
            "name": "Accessibility Test",
            "children": [
                {"role": "heading", "name": "Inputs", "level": 1},
                {"role": "textbox", "name": "Empty input", "focused": True},
                {"role": "textbox", "name": "readonly input", "readonly": True},
                {"role": "textbox", "name": "disabled input", "disabled": True},
                {"role": "textbox", "name": "Input with whitespace", "value": "  "},
                {"role": "textbox", "name": "", "value": "value only"},
                {
                    "role": "textbox",
                    "name": "",
                    "value": "and a value",
                },  # firefox doesn't use aria-placeholder for the name
                {
                    "role": "textbox",
                    "name": "",
                    "value": "and a value",
                    "description": "This is a description!",
                },  # and here
            ],
        }
    elif is_chromium:
        golden = {
            "role": "WebArea",
            "name": "Accessibility Test",
            "children": [
                {"role": "heading", "name": "Inputs", "level": 1},
                {"role": "textbox", "name": "Empty input", "focused": True},
                {"role": "textbox", "name": "readonly input", "readonly": True},
                {"role": "textbox", "name": "disabled input", "disabled": True},
                {"role": "textbox", "name": "Input with whitespace", "value": "  "},
                {"role": "textbox", "name": "", "value": "value only"},
                {"role": "textbox", "name": "placeholder", "value": "and a value"},
                {
                    "role": "textbox",
                    "name": "placeholder",
                    "value": "and a value",
                    "description": "This is a description!",
                },
            ],
        }
    else:
        golden = {
            "role": "WebArea",
            "name": "Accessibility Test",
            "children": [
                {"role": "heading", "name": "Inputs", "level": 1},
                {"role": "textbox", "name": "Empty input", "focused": True},
                {"role": "textbox", "name": "readonly input", "readonly": True},
                {"role": "textbox", "name": "disabled input", "disabled": True},
                {"role": "textbox", "name": "Input with whitespace", "value": "  "},
                {"role": "textbox", "name": "", "value": "value only"},
                {"role": "textbox", "name": "placeholder", "value": "and a value"},
                {
                    "role": "textbox",
                    "name": "This is a description!",
                    "value": "and a value",
                },  # webkit uses the description over placeholder for the name
            ],
        }
    assert await page.accessibility.snapshot() == golden


async def test_accessibility_should_work_with_regular_text(
    page: Page, is_firefox: bool
) -> None:
    await page.set_content("<div>Hello World</div>")
    snapshot = await page.accessibility.snapshot()
    assert snapshot
    assert snapshot["children"][0] == {
        "role": "text leaf" if is_firefox else "text",
        "name": "Hello World",
    }


async def test_accessibility_roledescription(page: Page) -> None:
    await page.set_content('<p tabIndex=-1 aria-roledescription="foo">Hi</p>')
    snapshot = await page.accessibility.snapshot()
    assert snapshot
    assert snapshot["children"][0]["roledescription"] == "foo"


async def test_accessibility_orientation(page: Page) -> None:
    await page.set_content(
        '<a href="" role="slider" aria-orientation="vertical">11</a>'
    )
    snapshot = await page.accessibility.snapshot()
    assert snapshot
    assert snapshot["children"][0]["orientation"] == "vertical"


async def test_accessibility_autocomplete(page: Page) -> None:
    await page.set_content('<div role="textbox" aria-autocomplete="list">hi</div>')
    snapshot = await page.accessibility.snapshot()
    assert snapshot
    assert snapshot["children"][0]["autocomplete"] == "list"


async def test_accessibility_multiselectable(page: Page) -> None:
    await page.set_content(
        '<div role="grid" tabIndex=-1 aria-multiselectable=true>hey</div>'
    )
    snapshot = await page.accessibility.snapshot()
    assert snapshot
    assert snapshot["children"][0]["multiselectable"]


async def test_accessibility_keyshortcuts(page: Page) -> None:
    await page.set_content(
        '<div role="grid" tabIndex=-1 aria-keyshortcuts="foo">hey</div>'
    )
    snapshot = await page.accessibility.snapshot()
    assert snapshot
    assert snapshot["children"][0]["keyshortcuts"] == "foo"


async def test_accessibility_filtering_children_of_leaf_nodes_should_not_report_text_nodes_inside_controls(
    page: Page, is_firefox: bool
) -> None:
    await page.set_content(
        """
    <div role="tablist">
    <div role="tab" aria-selected="true"><b>Tab1</b></div>
    <div role="tab">Tab2</div>
    </div>"""
    )
    golden = {
        "role": "document" if is_firefox else "WebArea",
        "name": "",
        "children": [
            {"role": "tab", "name": "Tab1", "selected": True},
            {"role": "tab", "name": "Tab2"},
        ],
    }
    assert await page.accessibility.snapshot() == golden


# Firefox does not support contenteditable="plaintext-only".
# WebKit rich text accessibility is iffy
@pytest.mark.only_browser("chromium")
async def test_accessibility_plain_text_field_with_role_should_not_have_children(
    page: Page,
) -> None:
    await page.set_content(
        """
    <div contenteditable="plaintext-only" role='textbox'>Edit this image:<img src="fakeimage.png" alt="my fake image"></div>"""
    )
    snapshot = await page.accessibility.snapshot()
    assert snapshot
    assert snapshot["children"][0] == {
        "multiline": True,
        "name": "",
        "role": "textbox",
        "value": "Edit this image:",
    }


@pytest.mark.only_browser("chromium")
async def test_accessibility_plain_text_field_without_role_should_not_have_content(
    page: Page,
) -> None:
    await page.set_content(
        """
    <div contenteditable="plaintext-only">Edit this image:<img src="fakeimage.png" alt="my fake image"></div>"""
    )
    snapshot = await page.accessibility.snapshot()
    assert snapshot
    assert snapshot["children"][0] == {
        "name": "",
        "role": "generic",
        "value": "Edit this image:",
    }


@pytest.mark.only_browser("chromium")
async def test_accessibility_plain_text_field_with_tabindex_and_without_role_should_not_have_content(
    page: Page,
) -> None:
    await page.set_content(
        """
    <div contenteditable="plaintext-only" tabIndex=0>Edit this image:<img src="fakeimage.png" alt="my fake image"></div>"""
    )
    snapshot = await page.accessibility.snapshot()
    assert snapshot
    assert snapshot["children"][0] == {
        "name": "",
        "role": "generic",
        "value": "Edit this image:",
    }


async def test_accessibility_non_editable_textbox_with_role_and_tabIndex_and_label_should_not_have_children(
    page: Page, is_chromium: bool, is_firefox: bool
) -> None:
    await page.set_content(
        """
      <div role="textbox" tabIndex=0 aria-checked="true" aria-label="my favorite textbox">
        this is the inner content
        <img alt="yo" src="fakeimg.png">
      </div>"""
    )
    if is_firefox:
        golden = {
            "role": "textbox",
            "name": "my favorite textbox",
            "value": "this is the inner content yo",
        }
    elif is_chromium:
        golden = {
            "role": "textbox",
            "name": "my favorite textbox",
            "value": "this is the inner content ",
        }
    else:
        golden = {
            "role": "textbox",
            "name": "my favorite textbox",
            "value": "this is the inner content  ",
        }
    snapshot = await page.accessibility.snapshot()
    assert snapshot
    assert snapshot["children"][0] == golden


async def test_accessibility_checkbox_with_and_tabIndex_and_label_should_not_have_children(
    page: Page,
) -> None:
    await page.set_content(
        """
    <div role="checkbox" tabIndex=0 aria-checked="true" aria-label="my favorite checkbox">
    this is the inner content
    <img alt="yo" src="fakeimg.png">
    </div>"""
    )
    golden = {"role": "checkbox", "name": "my favorite checkbox", "checked": True}
    snapshot = await page.accessibility.snapshot()
    assert snapshot
    assert snapshot["children"][0] == golden


async def test_accessibility_checkbox_without_label_should_not_have_children(
    page: Page, is_firefox: bool
) -> None:
    await page.set_content(
        """
      <div role="checkbox" aria-checked="true">
        this is the inner content
        <img alt="yo" src="fakeimg.png">
    </div>"""
    )
    golden = {
        "role": "checkbox",
        "name": "this is the inner content yo",
        "checked": True,
    }
    snapshot = await page.accessibility.snapshot()
    assert snapshot
    assert snapshot["children"][0] == golden


async def test_accessibility_should_work_a_button(page: Page) -> None:
    await page.set_content("<button>My Button</button>")

    button = await page.query_selector("button")
    assert await page.accessibility.snapshot(root=button) == {
        "role": "button",
        "name": "My Button",
    }


async def test_accessibility_should_work_an_input(page: Page) -> None:
    await page.set_content('<input title="My Input" value="My Value">')

    input = await page.query_selector("input")
    assert await page.accessibility.snapshot(root=input) == {
        "role": "textbox",
        "name": "My Input",
        "value": "My Value",
    }


async def test_accessibility_should_work_on_a_menu(page: Page) -> None:
    await page.set_content(
        """
        <div role="menu" title="My Menu">
        <div role="menuitem">First Item</div>
        <div role="menuitem">Second Item</div>
        <div role="menuitem">Third Item</div>
        </div>
    """
    )

    menu = await page.query_selector('div[role="menu"]')
    golden = {
        "role": "menu",
        "name": "My Menu",
        "children": [
            {"role": "menuitem", "name": "First Item"},
            {"role": "menuitem", "name": "Second Item"},
            {"role": "menuitem", "name": "Third Item"},
        ],
    }
    actual = await page.accessibility.snapshot(root=menu)
    assert actual
    # Different per browser channel
    if "orientation" in actual:
        del actual["orientation"]
    assert actual == golden


async def test_accessibility_should_return_null_when_the_element_is_no_longer_in_DOM(
    page: Page,
) -> None:
    await page.set_content("<button>My Button</button>")
    button = await page.query_selector("button")
    await page.eval_on_selector("button", "button => button.remove()")
    assert await page.accessibility.snapshot(root=button) is None


async def test_accessibility_should_show_uninteresting_nodes(page: Page) -> None:
    await page.set_content(
        """
        <div id="root" role="textbox">
        <div>
            hello
            <div>
            world
            </div>
        </div>
        </div>
    """
    )

    root = await page.query_selector("#root")
    snapshot = await page.accessibility.snapshot(root=root, interesting_only=False)
    assert snapshot
    assert snapshot["role"] == "textbox"
    assert "hello" in snapshot["value"]
    assert "world" in snapshot["value"]
    assert snapshot["children"]



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import re
from pathlib import Path
from typing import AsyncContextManager, Callable

from playwright.async_api import (
    Browser,
    BrowserContext,
    BrowserType,
    Page,
    Response,
    expect,
)
from tests.server import Server

from .conftest import TraceViewerPage


async def test_browser_context_output_trace(
    browser: Browser, server: Server, tmp_path: Path
) -> None:
    context = await browser.new_context()
    await context.tracing.start(screenshots=True, snapshots=True)
    page = await context.new_page()
    await page.goto(server.PREFIX + "/grid.html")
    await context.tracing.stop(path=tmp_path / "trace.zip")
    assert Path(tmp_path / "trace.zip").exists()


async def test_start_stop(browser: Browser) -> None:
    context = await browser.new_context()
    await context.tracing.start()
    await context.tracing.stop()
    await context.close()


async def test_browser_context_should_not_throw_when_stopping_without_start_but_not_exporting(
    context: BrowserContext,
) -> None:
    await context.tracing.stop()


async def test_browser_context_output_trace_chunk(
    browser: Browser, server: Server, tmp_path: Path
) -> None:
    context = await browser.new_context()
    await context.tracing.start(screenshots=True, snapshots=True)
    page = await context.new_page()
    await page.goto(server.PREFIX + "/grid.html")
    button = page.locator(".box").first

    await context.tracing.start_chunk(title="foo")
    await button.click()
    await context.tracing.stop_chunk(path=tmp_path / "trace1.zip")
    assert Path(tmp_path / "trace1.zip").exists()

    await context.tracing.start_chunk(title="foo")
    await button.click()
    await context.tracing.stop_chunk(path=tmp_path / "trace2.zip")
    assert Path(tmp_path / "trace2.zip").exists()


async def test_should_collect_sources(
    context: BrowserContext,
    page: Page,
    server: Server,
    tmp_path: Path,
    show_trace_viewer: Callable[[Path], AsyncContextManager[TraceViewerPage]],
) -> None:
    await context.tracing.start(sources=True)
    await page.goto(server.EMPTY_PAGE)
    await page.set_content("<button>Click</button>")

    async def my_method_outer() -> None:
        async def my_method_inner() -> None:
            await page.get_by_text("Click").click()

        await my_method_inner()

    await my_method_outer()
    path = tmp_path / "trace.zip"
    await context.tracing.stop(path=path)

    async with show_trace_viewer(path) as trace_viewer:
        await expect(trace_viewer.action_titles).to_have_text(
            [
                re.compile(r'Navigate to "/empty\.html"'),
                re.compile(r"Set content"),
                re.compile(r"Click"),
            ]
        )
        await trace_viewer.show_source_tab()
        # Check that stack frames are shown (they might be anonymous in Python)
        await expect(trace_viewer.stack_frames).to_contain_text(
            [
                re.compile(r"my_method_inner"),
                re.compile(r"my_method_outer"),
                re.compile(r"test_should_collect_sources"),
            ]
        )

        await trace_viewer.select_action("Set content")
        # Check that the source file is shown
        await expect(
            trace_viewer.page.locator(".source-tab-file-name")
        ).to_have_attribute("title", re.compile(r".*test_.*\.py"))
        await expect(trace_viewer.page.locator(".source-line-running")).to_contain_text(
            'page.set_content("<button>Click</button>")'
        )


async def test_should_collect_trace_with_resources_but_no_js(
    context: BrowserContext,
    page: Page,
    server: Server,
    tmp_path: Path,
    show_trace_viewer: Callable[[Path], AsyncContextManager[TraceViewerPage]],
) -> None:
    await context.tracing.start(screenshots=True, snapshots=True)
    await page.goto(server.PREFIX + "/frames/frame.html")
    await page.set_content("<button>Click</button>")
    await page.click('"Click"')
    await page.mouse.move(20, 20)
    await page.mouse.dblclick(30, 30)
    await page.request.get(server.EMPTY_PAGE)
    await page.keyboard.insert_text("abc")
    await page.wait_for_timeout(2000)  # Give it some time to produce screenshots.
    await page.route("**/empty.html", lambda route: route.continue_())
    await page.goto(server.EMPTY_PAGE)
    await page.goto(server.PREFIX + "/one-style.html")
    await page.close()
    trace_file_path = tmp_path / "trace.zip"
    await context.tracing.stop(path=trace_file_path)

    async with show_trace_viewer(trace_file_path) as trace_viewer:
        await expect(trace_viewer.action_titles).to_have_text(
            [
                re.compile(r'Navigate to "/frames/frame\.html"'),
                re.compile(r"Set content"),
                re.compile(r"Click"),
                re.compile(r"Mouse move"),
                re.compile(r"Double click"),
                re.compile(r"GET \"/empty\.html\""),
                re.compile(r'Insert "abc"'),
                re.compile(r"Wait for timeout"),
                re.compile(r'Navigate to "/empty\.html"'),
                re.compile(r'Navigate to "/one-style\.html"'),
                re.compile(r"Close"),
            ]
        )

        await trace_viewer.select_action("Set content")
        await expect(
            trace_viewer.page.locator(".browser-frame-address-bar")
        ).to_have_text(server.PREFIX + "/frames/frame.html")
        frame = await trace_viewer.snapshot_frame("Set content", 0, False)
        await expect(frame.locator("button")).to_have_text("Click")


async def test_should_correctly_determine_sync_apiname(
    context: BrowserContext,
    page: Page,
    server: Server,
    tmp_path: Path,
    show_trace_viewer: Callable,
) -> None:
    await context.tracing.start(screenshots=True, snapshots=True)

    received_response: "asyncio.Future[None]" = asyncio.Future()

    async def _handle_response(response: Response) -> None:
        await response.request.all_headers()
        await response.text()
        received_response.set_result(None)

    page.once("response", _handle_response)
    await page.goto(server.PREFIX + "/grid.html")
    await received_response
    await page.close()
    trace_file_path = tmp_path / "trace.zip"
    await context.tracing.stop(path=trace_file_path)

    async with show_trace_viewer(trace_file_path) as trace_viewer:
        await expect(trace_viewer.action_titles).to_have_text(
            [
                re.compile(r'Navigate to "/grid\.html"'),
                re.compile(r"Close"),
            ]
        )


async def test_should_collect_two_traces(
    context: BrowserContext,
    page: Page,
    server: Server,
    tmp_path: Path,
    show_trace_viewer: Callable[[Path], AsyncContextManager[TraceViewerPage]],
) -> None:
    await context.tracing.start(screenshots=True, snapshots=True)
    await page.goto(server.EMPTY_PAGE)
    await page.set_content("<button>Click</button>")
    await page.click('"Click"')
    tracing1_path = tmp_path / "trace1.zip"
    await context.tracing.stop(path=tracing1_path)

    await context.tracing.start(screenshots=True, snapshots=True)
    await page.dblclick('"Click"')
    await page.close()
    tracing2_path = tmp_path / "trace2.zip"
    await context.tracing.stop(path=tracing2_path)

    async with show_trace_viewer(tracing1_path) as trace_viewer:
        await expect(trace_viewer.action_titles).to_have_text(
            [
                re.compile(r'Navigate to "/empty\.html"'),
                re.compile(r"Set content"),
                re.compile(r"Click"),
            ]
        )

    async with show_trace_viewer(tracing2_path) as trace_viewer:
        await expect(trace_viewer.action_titles).to_have_text(
            [
                re.compile(r"Double click"),
                re.compile(r"Close"),
            ]
        )


async def test_should_work_with_playwright_context_managers(
    context: BrowserContext,
    page: Page,
    server: Server,
    tmp_path: Path,
    show_trace_viewer: Callable[[Path], AsyncContextManager[TraceViewerPage]],
) -> None:
    await context.tracing.start(screenshots=True, snapshots=True)
    await page.goto(server.EMPTY_PAGE)
    await page.set_content("<button>Click</button>")
    async with page.expect_console_message() as message_info:
        await page.evaluate('() => console.log("hello")')
        await page.click('"Click"')
    assert (await message_info.value).text == "hello"

    async with page.expect_popup():
        await page.evaluate("window._popup = window.open(document.location.href)")
    trace_file_path = tmp_path / "trace.zip"
    await context.tracing.stop(path=trace_file_path)

    async with show_trace_viewer(trace_file_path) as trace_viewer:
        await expect(trace_viewer.action_titles).to_have_text(
            [
                re.compile(r'Navigate to "/empty\.html"'),
                re.compile(r"Set content"),
                re.compile(r'Wait for event "page\.expect_event\(console\)"'),
                re.compile(r"Evaluate"),
                re.compile(r"Click"),
                re.compile(r'Wait for event "page\.expect_event\(popup\)"'),
                re.compile(r"Evaluate"),
            ]
        )


async def test_should_display_wait_for_load_state_even_if_did_not_wait_for_it(
    context: BrowserContext,
    page: Page,
    server: Server,
    tmp_path: Path,
    show_trace_viewer: Callable[[Path], AsyncContextManager[TraceViewerPage]],
) -> None:
    await context.tracing.start(screenshots=True, snapshots=True)

    await page.goto(server.EMPTY_PAGE)
    await page.wait_for_load_state("load")
    await page.wait_for_load_state("load")

    trace_file_path = tmp_path / "trace.zip"
    await context.tracing.stop(path=trace_file_path)

    async with show_trace_viewer(trace_file_path) as trace_viewer:
        await expect(trace_viewer.action_titles).to_have_text(
            [
                re.compile(r'Navigate to "/empty\.html"'),
                re.compile(r'Wait for event "frame\.wait_for_load_state"'),
                re.compile(r'Wait for event "frame\.wait_for_load_state"'),
            ]
        )


async def test_should_respect_traces_dir_and_name(
    browser_type: BrowserType,
    server: Server,
    tmp_path: Path,
    launch_arguments: dict,
    show_trace_viewer: Callable[[Path], AsyncContextManager[TraceViewerPage]],
) -> None:
    traces_dir = tmp_path / "traces"
    browser = await browser_type.launch(traces_dir=traces_dir, **launch_arguments)
    context = await browser.new_context()
    page = await context.new_page()

    await context.tracing.start(name="name1", snapshots=True)
    await page.goto(server.PREFIX + "/one-style.html")
    await context.tracing.stop_chunk(path=tmp_path / "trace1.zip")
    assert (traces_dir / "name1.trace").exists()
    assert (traces_dir / "name1.network").exists()

    await context.tracing.start_chunk(name="name2")
    await page.goto(server.PREFIX + "/har.html")
    await context.tracing.stop(path=tmp_path / "trace2.zip")
    assert (traces_dir / "name2.trace").exists()
    assert (traces_dir / "name2.network").exists()

    await browser.close()

    async with show_trace_viewer(tmp_path / "trace1.zip") as trace_viewer:
        await expect(trace_viewer.action_titles).to_have_text(
            [
                re.compile('Navigate to "/one-style\\.html"'),
            ]
        )
        frame = await trace_viewer.snapshot_frame(
            'Navigate to "/one-style.html"', 0, False
        )
        await expect(frame.locator("body")).to_have_css(
            "background-color", "rgb(255, 192, 203)"
        )
        await expect(frame.locator("body")).to_have_text("hello, world!")

    async with show_trace_viewer(tmp_path / "trace2.zip") as trace_viewer:
        await expect(trace_viewer.action_titles).to_have_text(
            [
                re.compile(r'Navigate to "/har\.html"'),
            ]
        )
        frame = await trace_viewer.snapshot_frame('Navigate to "/har.html"', 0, False)
        await expect(frame.locator("body")).to_have_css(
            "background-color", "rgb(255, 192, 203)"
        )
        await expect(frame.locator("body")).to_have_text("hello, world!")


async def test_should_show_tracing_group_in_action_list(
    context: BrowserContext,
    tmp_path: Path,
    show_trace_viewer: Callable[[Path], AsyncContextManager[TraceViewerPage]],
) -> None:
    await context.tracing.start()
    page = await context.new_page()

    await context.tracing.group("outer group")
    await page.goto("data:text/html,<!DOCTYPE html><body><div>Hello world</div></body>")
    await context.tracing.group("inner group 1")
    await page.locator("body").click()
    await context.tracing.group_end()
    await context.tracing.group("inner group 2")
    await page.get_by_text("Hello").is_visible()
    await context.tracing.group_end()
    await context.tracing.group_end()

    trace_path = tmp_path / "trace.zip"
    await context.tracing.stop(path=trace_path)

    async with show_trace_viewer(trace_path) as trace_viewer:
        await trace_viewer.expand_action("inner group 1")
        await expect(trace_viewer.action_titles).to_have_text(
            [
                re.compile(r"Create page"),
                re.compile(r"outer group"),
                re.compile(r"Navigate to \"data:\""),
                re.compile(r"inner group 1"),
                re.compile(r"Click"),
                re.compile(r"inner group 2"),
            ]
        )



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
from asyncio import Future
from pathlib import Path
from typing import Dict, List, Optional, Union

import pytest
from twisted.web import http

from playwright.async_api import Browser, Error, Page, Request, Response, Route
from tests.server import Server, TestServerRequest

from .utils import Utils


def adjust_server_headers(headers: Dict[str, str], browser_name: str) -> Dict[str, str]:
    if browser_name != "firefox":
        return headers
    headers = headers.copy()
    headers.pop("priority", None)
    return headers


async def test_request_fulfill(page: Page, server: Server) -> None:
    async def handle_request(route: Route, request: Request) -> None:
        headers = await route.request.all_headers()
        assert headers["accept"]
        assert route.request == request
        assert repr(route) == f"<Route request={route.request}>"
        assert "empty.html" in request.url
        assert request.headers["user-agent"]
        assert request.method == "GET"
        assert request.post_data is None
        assert request.is_navigation_request()
        assert request.resource_type == "document"
        assert request.frame == page.main_frame
        assert request.frame.url == "about:blank"
        assert (
            repr(request) == f"<Request url={request.url!r} method={request.method!r}>"
        )
        await route.fulfill(body="Text")

    await page.route(
        "**/empty.html",
        lambda route, request: asyncio.create_task(handle_request(route, request)),
    )

    response = await page.goto(server.EMPTY_PAGE)
    assert response

    assert response.ok
    assert (
        repr(response) == f"<Response url={response.url!r} request={response.request}>"
    )
    assert await response.text() == "Text"


async def test_request_continue(page: Page, server: Server) -> None:
    async def handle_request(
        route: Route, request: Request, intercepted: List[bool]
    ) -> None:
        intercepted.append(True)
        await route.continue_()

    intercepted: List[bool] = []
    await page.route(
        "**/*",
        lambda route, request: asyncio.create_task(
            handle_request(route, request, intercepted)
        ),
    )

    response = await page.goto(server.EMPTY_PAGE)
    assert response
    assert response.ok
    assert intercepted == [True]
    assert await page.title() == ""


async def test_page_events_request_should_fire_for_navigation_requests(
    page: Page, server: Server
) -> None:
    requests = []
    page.on("request", lambda r: requests.append(r))
    await page.goto(server.EMPTY_PAGE)
    assert len(requests) == 1


async def test_page_events_request_should_accept_method(
    page: Page, server: Server
) -> None:
    class Log:
        def __init__(self) -> None:
            self.requests: List[Request] = []

        def handle(self, request: Request) -> None:
            self.requests.append(request)

    log = Log()
    page.on("request", log.handle)
    await page.goto(server.EMPTY_PAGE)
    assert len(log.requests) == 1


async def test_page_events_request_should_fire_for_iframes(
    page: Page, server: Server, utils: Utils
) -> None:
    requests = []
    page.on("request", lambda r: requests.append(r))
    await page.goto(server.EMPTY_PAGE)
    await utils.attach_frame(page, "frame1", server.EMPTY_PAGE)
    assert len(requests) == 2


async def test_page_events_request_should_fire_for_fetches(
    page: Page, server: Server
) -> None:
    requests = []
    page.on("request", lambda r: requests.append(r))
    await page.goto(server.EMPTY_PAGE)
    await page.evaluate('() => fetch("/empty.html")')
    assert len(requests) == 2


async def test_page_events_request_should_report_requests_and_responses_handled_by_service_worker(
    page: Page, server: Server
) -> None:
    await page.goto(server.PREFIX + "/serviceworkers/fetchdummy/sw.html")
    await page.evaluate("() => window.activationPromise")
    sw_response = None
    async with page.expect_request("**/*") as request_info:
        sw_response = await page.evaluate('() => fetchDummy("foo")')
    request = await request_info.value
    assert sw_response == "responseFromServiceWorker:foo"
    assert request.url == server.PREFIX + "/serviceworkers/fetchdummy/foo"
    response = await request.response()
    assert response
    assert response.url == server.PREFIX + "/serviceworkers/fetchdummy/foo"
    assert await response.text() == "responseFromServiceWorker:foo"


async def test_request_frame_should_work_for_main_frame_navigation_request(
    page: Page, server: Server
) -> None:
    requests = []
    page.on("request", lambda r: requests.append(r))
    await page.goto(server.EMPTY_PAGE)
    assert len(requests) == 1
    assert requests[0].frame == page.main_frame


async def test_request_frame_should_work_for_subframe_navigation_request(
    page: Page, server: Server, utils: Utils
) -> None:
    await page.goto(server.EMPTY_PAGE)
    requests = []
    page.on("request", lambda r: requests.append(r))
    await utils.attach_frame(page, "frame1", server.EMPTY_PAGE)
    assert len(requests) == 1
    assert requests[0].frame == page.frames[1]


async def test_request_frame_should_work_for_fetch_requests(
    page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    requests: List[Request] = []
    page.on("request", lambda r: requests.append(r))
    await page.evaluate('() => fetch("/digits/1.png")')
    requests = [r for r in requests if "favicon" not in r.url]
    assert len(requests) == 1
    assert requests[0].frame == page.main_frame


async def test_request_headers_should_work(
    page: Page, server: Server, is_chromium: bool, is_firefox: bool, is_webkit: bool
) -> None:
    response = await page.goto(server.EMPTY_PAGE)
    assert response
    if is_chromium:
        assert "Chrome" in response.request.headers["user-agent"]
    elif is_firefox:
        assert "Firefox" in response.request.headers["user-agent"]
    elif is_webkit:
        assert "WebKit" in response.request.headers["user-agent"]


async def test_request_headers_should_get_the_same_headers_as_the_server(
    page: Page,
    server: Server,
    is_webkit: bool,
    is_win: bool,
    browser_name: str,
) -> None:
    if is_webkit and is_win:
        pytest.xfail("Curl does not show accept-encoding and accept-language")
    server_request_headers_future: Future[Dict[str, str]] = asyncio.Future()

    def handle(request: http.Request) -> None:
        normalized_headers = {
            key.decode().lower(): value[0].decode()
            for key, value in request.requestHeaders.getAllRawHeaders()
        }
        server_request_headers_future.set_result(normalized_headers)
        request.write(b"done")
        request.finish()

    server.set_route("/empty.html", handle)
    response = await page.goto(server.EMPTY_PAGE)
    assert response
    server_headers = adjust_server_headers(
        await server_request_headers_future, browser_name
    )
    assert await response.request.all_headers() == server_headers


async def test_request_headers_should_get_the_same_headers_as_the_server_cors(
    page: Page, server: Server, is_webkit: bool, is_win: bool, browser_name: str
) -> None:
    if is_webkit and is_win:
        pytest.xfail("Curl does not show accept-encoding and accept-language")
    await page.goto(server.PREFIX + "/empty.html")
    server_request_headers_future: Future[Dict[str, str]] = asyncio.Future()

    def handle_something(request: http.Request) -> None:
        normalized_headers = {
            key.decode().lower(): value[0].decode()
            for key, value in request.requestHeaders.getAllRawHeaders()
        }
        server_request_headers_future.set_result(normalized_headers)
        request.setHeader("Access-Control-Allow-Origin", "*")
        request.write(b"done")
        request.finish()

    server.set_route("/something", handle_something)

    text = None
    async with page.expect_request("**/*") as request_info:
        text = await page.evaluate(
            """async url => {
                const data = await fetch(url);
                return data.text();
            }""",
            server.CROSS_PROCESS_PREFIX + "/something",
        )
    request = await request_info.value
    assert text == "done"
    server_headers = adjust_server_headers(
        await server_request_headers_future, browser_name
    )
    assert await request.all_headers() == server_headers


async def test_should_report_request_headers_array(
    page: Page, server: Server, is_win: bool, browser_name: str
) -> None:
    if is_win and browser_name == "webkit":
        pytest.skip("libcurl does not support non-set-cookie multivalue headers")
    expected_headers = []

    def handle(request: http.Request) -> None:
        for name, values in request.requestHeaders.getAllRawHeaders():
            for value in values:
                if browser_name == "firefox" and name.decode().lower() == "priority":
                    continue
                expected_headers.append(
                    {"name": name.decode().lower(), "value": value.decode()}
                )
        request.finish()

    server.set_route("/headers", handle)
    await page.goto(server.EMPTY_PAGE)
    async with page.expect_request("*/**") as request_info:
        await page.evaluate(
            """() => fetch('/headers', {
            headers: [
                ['header-a', 'value-a'],
                ['header-b', 'value-b'],
                ['header-a', 'value-a-1'],
                ['header-a', 'value-a-2'],
            ]
            })
        """
        )
    request = await request_info.value
    sorted_pw_request_headers = sorted(
        list(
            map(
                lambda header: {
                    "name": header["name"].lower(),
                    "value": header["value"],
                },
                await request.headers_array(),
            )
        ),
        key=lambda header: header["name"],
    )
    sorted_expected_headers = sorted(
        expected_headers, key=lambda header: header["name"]
    )
    assert sorted_pw_request_headers == sorted_expected_headers
    assert await request.header_value("Header-A") == "value-a, value-a-1, value-a-2"
    assert await request.header_value("not-there") is None


async def test_should_report_response_headers_array(
    page: Page, server: Server, is_win: bool, browser_name: str
) -> None:
    if is_win and browser_name == "webkit":
        pytest.skip("libcurl does not support non-set-cookie multivalue headers")
    expected_headers = {
        "header-a": ["value-a", "value-a-1", "value-a-2"],
        "header-b": ["value-b"],
        "set-cookie": ["a=b", "c=d"],
    }

    def handle(request: http.Request) -> None:
        for key in expected_headers:
            for value in expected_headers[key]:
                request.responseHeaders.addRawHeader(key, value)
        request.finish()

    server.set_route("/headers", handle)
    await page.goto(server.EMPTY_PAGE)
    async with page.expect_response("*/**") as response_info:
        await page.evaluate(
            """() => fetch('/headers')
        """
        )
    response = await response_info.value
    actual_headers: Dict[str, List[str]] = {}
    for header in await response.headers_array():
        name = header["name"].lower()
        value = header["value"]
        if not actual_headers.get(name):
            actual_headers[name] = []
        actual_headers[name].append(value)

    for key in ["Keep-Alive", "Connection", "Date", "Transfer-Encoding"]:
        if key in actual_headers:
            actual_headers.pop(key)
        if key.lower() in actual_headers:
            actual_headers.pop(key.lower())
    assert actual_headers == expected_headers
    assert await response.header_value("not-there") is None
    assert await response.header_value("set-cookie") == "a=b\nc=d"
    assert await response.header_value("header-a") == "value-a, value-a-1, value-a-2"
    assert await response.header_values("set-cookie") == ["a=b", "c=d"]


async def test_response_headers_should_work(page: Page, server: Server) -> None:
    server.set_route("/empty.html", lambda r: (r.setHeader("foo", "bar"), r.finish()))

    response = await page.goto(server.EMPTY_PAGE)
    assert response
    assert response.headers["foo"] == "bar"
    assert (await response.all_headers())["foo"] == "bar"


async def test_request_post_data_should_work(page: Page, server: Server) -> None:
    await page.goto(server.EMPTY_PAGE)
    server.set_route("/post", lambda r: r.finish())
    requests = []
    page.on("request", lambda r: requests.append(r))
    await page.evaluate(
        '() => fetch("./post", { method: "POST", body: JSON.stringify({foo: "bar"})})'
    )
    assert len(requests) == 1
    assert requests[0].post_data == '{"foo":"bar"}'


async def test_request_post_data__should_be_undefined_when_there_is_no_post_data(
    page: Page, server: Server
) -> None:
    response = await page.goto(server.EMPTY_PAGE)
    assert response
    assert response.request.post_data is None


async def test_should_parse_the_json_post_data(page: Page, server: Server) -> None:
    await page.goto(server.EMPTY_PAGE)
    server.set_route("/post", lambda req: req.finish())
    requests = []
    page.on("request", lambda r: requests.append(r))
    await page.evaluate(
        """() => fetch('./post', { method: 'POST', body: JSON.stringify({ foo: 'bar' }) })"""
    )
    assert len(requests) == 1
    assert requests[0].post_data_json == {"foo": "bar"}


async def test_should_parse_the_data_if_content_type_is_form_urlencoded(
    page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    server.set_route("/post", lambda req: req.finish())
    requests = []
    page.on("request", lambda r: requests.append(r))
    await page.set_content(
        """<form method='POST' action='/post'><input type='text' name='foo' value='bar'><input type='number' name='baz' value='123'><input type='submit'></form>"""
    )
    await page.click("input[type=submit]")
    assert len(requests) == 1
    assert requests[0].post_data_json == {"foo": "bar", "baz": "123"}


async def test_should_be_undefined_when_there_is_no_post_data(
    page: Page, server: Server
) -> None:
    response = await page.goto(server.EMPTY_PAGE)
    assert response
    assert response.request.post_data_json is None


async def test_should_return_post_data_without_content_type(
    page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    async with page.expect_request("**/*") as request_info:
        await page.evaluate(
            """({url}) => {
            const request = new Request(url, {
                method: 'POST',
                body: JSON.stringify({ value: 42 }),
            });
            request.headers.set('content-type', '');
            return fetch(request);
        }""",
            {"url": server.PREFIX + "/title.html"},
        )
    request = await request_info.value
    assert request.post_data_json == {"value": 42}


async def test_should_throw_on_invalid_json_in_post_data(
    page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    async with page.expect_request("**/*") as request_info:
        await page.evaluate(
            """({url}) => {
            const request = new Request(url, {
                method: 'POST',
                body: '<not a json>',
            });
            request.headers.set('content-type', '');
            return fetch(request);
        }""",
            {"url": server.PREFIX + "/title.html"},
        )
    request = await request_info.value
    with pytest.raises(Error) as exc_info:
        print(request.post_data_json)
    assert "POST data is not a valid JSON object: <not a json>" in str(exc_info.value)


async def test_should_work_with_binary_post_data(page: Page, server: Server) -> None:
    await page.goto(server.EMPTY_PAGE)
    server.set_route("/post", lambda req: req.finish())
    requests = []
    page.on("request", lambda r: requests.append(r))
    await page.evaluate(
        """async () => {
        await fetch('./post', { method: 'POST', body: new Uint8Array(Array.from(Array(256).keys())) })
    }"""
    )
    assert len(requests) == 1
    buffer = requests[0].post_data_buffer
    assert len(buffer) == 256
    for i in range(256):
        assert buffer[i] == i


async def test_should_work_with_binary_post_data_and_interception(
    page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    server.set_route("/post", lambda req: req.finish())
    requests = []
    await page.route("/post", lambda route: asyncio.ensure_future(route.continue_()))
    page.on("request", lambda r: requests.append(r))
    await page.evaluate(
        """async () => {
        await fetch('./post', { method: 'POST', body: new Uint8Array(Array.from(Array(256).keys())) })
    }"""
    )
    assert len(requests) == 1
    buffer = requests[0].post_data_buffer
    assert len(buffer) == 256
    for i in range(256):
        assert buffer[i] == i


async def test_response_text_should_work(page: Page, server: Server) -> None:
    response = await page.goto(server.PREFIX + "/simple.json")
    assert response
    assert await response.text() == '{"foo": "bar"}\n'


async def test_response_text_should_return_uncompressed_text(
    page: Page, server: Server
) -> None:
    server.enable_gzip("/simple.json")
    response = await page.goto(server.PREFIX + "/simple.json")
    assert response
    assert response.headers["content-encoding"] == "gzip"
    assert await response.text() == '{"foo": "bar"}\n'


async def test_response_text_should_throw_when_requesting_body_of_redirected_response(
    page: Page, server: Server
) -> None:
    server.set_redirect("/foo.html", "/empty.html")
    response = await page.goto(server.PREFIX + "/foo.html")
    assert response
    redirected_from = response.request.redirected_from
    assert redirected_from
    redirected = await redirected_from.response()
    assert redirected
    assert redirected.status == 302
    error: Optional[Error] = None
    try:
        await redirected.text()
    except Error as exc:
        error = exc
    assert error
    assert "Response body is unavailable for redirect responses" in error.message


async def test_response_json_should_work(page: Page, server: Server) -> None:
    response = await page.goto(server.PREFIX + "/simple.json")
    assert response
    assert await response.json() == {"foo": "bar"}


async def test_response_body_should_work(
    page: Page, server: Server, assetdir: Path
) -> None:
    response = await page.goto(server.PREFIX + "/pptr.png")
    assert response
    with open(
        assetdir / "pptr.png",
        "rb",
    ) as fd:
        assert fd.read() == await response.body()


async def test_response_body_should_work_with_compression(
    page: Page, server: Server, assetdir: Path
) -> None:
    server.enable_gzip("/pptr.png")
    response = await page.goto(server.PREFIX + "/pptr.png")
    assert response
    with open(
        assetdir / "pptr.png",
        "rb",
    ) as fd:
        assert fd.read() == await response.body()


async def test_response_status_text_should_work(page: Page, server: Server) -> None:
    server.set_route("/cool", lambda r: (r.setResponseCode(200, b"cool!"), r.finish()))

    response = await page.goto(server.PREFIX + "/cool")
    assert response
    assert response.status_text == "cool!"


async def test_request_resource_type_should_return_event_source(
    page: Page, server: Server
) -> None:
    SSE_MESSAGE = {"foo": "bar"}
    # 1. Setup server-sent events on server that immediately sends a message to the client.
    server.set_route(
        "/sse",
        lambda r: (
            r.setHeader("Content-Type", "text/event-stream"),
            r.setHeader("Connection", "keep-alive"),
            r.setHeader("Cache-Control", "no-cache"),
            r.setResponseCode(200),
            r.write(f"data: {json.dumps(SSE_MESSAGE)}\n\n".encode()),
            r.finish(),
        ),
    )

    # 2. Subscribe to page request events.
    await page.goto(server.EMPTY_PAGE)
    requests = []
    page.on("request", lambda r: requests.append(r))
    # 3. Connect to EventSource in browser and return first message.
    assert (
        await page.evaluate(
            """() => {
      const eventSource = new EventSource('/sse');
      return new Promise(resolve => {
        eventSource.onmessage = e => resolve(JSON.parse(e.data));
      });
    }"""
        )
        == SSE_MESSAGE
    )
    assert requests[0].resource_type == "eventsource"


async def test_network_events_request(page: Page, server: Server) -> None:
    requests = []
    page.on("request", lambda r: requests.append(r))
    await page.goto(server.EMPTY_PAGE)
    assert len(requests) == 1
    assert requests[0].url == server.EMPTY_PAGE
    assert requests[0].resource_type == "document"
    assert requests[0].method == "GET"
    assert await requests[0].response()
    assert requests[0].frame == page.main_frame
    assert requests[0].frame.url == server.EMPTY_PAGE


async def test_network_events_response(page: Page, server: Server) -> None:
    responses = []
    page.on("response", lambda r: responses.append(r))
    await page.goto(server.EMPTY_PAGE)
    assert len(responses) == 1
    assert responses[0].url == server.EMPTY_PAGE
    assert responses[0].status == 200
    assert responses[0].ok
    assert responses[0].request


async def test_network_events_request_failed(
    page: Page,
    server: Server,
    is_chromium: bool,
    is_webkit: bool,
    is_mac: bool,
    is_win: bool,
) -> None:
    def handle_request(request: TestServerRequest) -> None:
        request.setHeader("Content-Type", "text/css")
        request.loseConnection()

    server.set_route("/one-style.css", handle_request)

    failed_requests = []
    page.on("requestfailed", lambda request: failed_requests.append(request))
    await page.goto(server.PREFIX + "/one-style.html")
    # TODO: https://github.com/microsoft/playwright/issues/12789
    assert len(failed_requests) >= 1
    assert "one-style.css" in failed_requests[0].url
    assert await failed_requests[0].response() is None
    assert failed_requests[0].resource_type == "stylesheet"
    if is_chromium:
        assert failed_requests[0].failure == "net::ERR_EMPTY_RESPONSE"
    elif is_webkit:
        if is_mac:
            assert failed_requests[0].failure == "The network connection was lost."
        elif is_win:
            assert (
                failed_requests[0].failure
                == "Server returned nothing (no headers, no data)"
            )
        else:
            assert failed_requests[0].failure in [
                "Message Corrupt",
                "Connection terminated unexpectedly",
            ]
    else:
        assert failed_requests[0].failure == "NS_ERROR_NET_RESET"
    assert failed_requests[0].frame


async def test_network_events_request_finished(page: Page, server: Server) -> None:
    async with page.expect_event("requestfinished") as event_info:
        await page.goto(server.EMPTY_PAGE)
    request = await event_info.value
    assert request.url == server.EMPTY_PAGE
    assert await request.response()
    assert request.frame == page.main_frame
    assert request.frame.url == server.EMPTY_PAGE


async def test_network_events_should_fire_events_in_proper_order(
    page: Page, server: Server
) -> None:
    events = []
    page.on("request", lambda request: events.append("request"))
    page.on("response", lambda response: events.append("response"))
    response = await page.goto(server.EMPTY_PAGE)
    assert response
    await response.finished()
    events.append("requestfinished")
    assert events == ["request", "response", "requestfinished"]


async def test_network_events_should_support_redirects(
    page: Page, server: Server
) -> None:
    FOO_URL = server.PREFIX + "/foo.html"
    events: Dict[str, List[Union[str, int]]] = {}
    events[FOO_URL] = []
    events[server.EMPTY_PAGE] = []

    def _handle_on_request(request: Request) -> None:
        events[request.url].append(request.method)

    page.on("request", _handle_on_request)

    def _handle_on_response(response: Response) -> None:
        events[response.url].append(response.status)

    page.on("response", _handle_on_response)

    def _handle_on_requestfinished(request: Request) -> None:
        events[request.url].append("DONE")

    page.on("requestfinished", _handle_on_requestfinished)

    def _handle_on_requestfailed(request: Request) -> None:
        events[request.url].append("FAIL")

    page.on("requestfailed", _handle_on_requestfailed)
    server.set_redirect("/foo.html", "/empty.html")
    response = await page.goto(FOO_URL)
    assert response
    await response.finished()
    expected = {}
    expected[FOO_URL] = ["GET", 302, "DONE"]
    expected[server.EMPTY_PAGE] = ["GET", 200, "DONE"]
    assert events == expected
    redirected_from = response.request.redirected_from
    assert redirected_from
    assert "/foo.html" in redirected_from.url
    assert redirected_from.redirected_from is None
    assert redirected_from.redirected_to == response.request


async def test_request_is_navigation_request_should_work(
    page: Page, server: Server
) -> None:
    requests: Dict[str, Request] = {}

    def handle_request(request: Request) -> None:
        requests[request.url.split("/").pop()] = request

    page.on("request", handle_request)
    server.set_redirect("/rrredirect", "/frames/one-frame.html")
    await page.goto(server.PREFIX + "/rrredirect")
    assert requests["rrredirect"].is_navigation_request()
    assert requests["one-frame.html"].is_navigation_request()
    assert requests["frame.html"].is_navigation_request()
    assert requests["script.js"].is_navigation_request() is False
    assert requests["style.css"].is_navigation_request() is False


async def test_request_is_navigation_request_should_work_when_navigating_to_image(
    page: Page, server: Server
) -> None:
    requests = []
    page.on("request", lambda r: requests.append(r))
    await page.goto(server.PREFIX + "/pptr.png")
    assert requests[0].is_navigation_request()


async def test_set_extra_http_headers_should_work(page: Page, server: Server) -> None:
    await page.set_extra_http_headers({"foo": "bar"})

    request = (
        await asyncio.gather(
            server.wait_for_request("/empty.html"),
            page.goto(server.EMPTY_PAGE),
        )
    )[0]
    assert request.getHeader("foo") == "bar"


async def test_set_extra_http_headers_should_work_with_redirects(
    page: Page, server: Server
) -> None:
    server.set_redirect("/foo.html", "/empty.html")
    await page.set_extra_http_headers({"foo": "bar"})

    request = (
        await asyncio.gather(
            server.wait_for_request("/empty.html"),
            page.goto(server.PREFIX + "/foo.html"),
        )
    )[0]
    assert request.getHeader("foo") == "bar"


async def test_set_extra_http_headers_should_work_with_extra_headers_from_browser_context(
    browser: Browser, server: Server
) -> None:
    context = await browser.new_context()
    await context.set_extra_http_headers({"foo": "bar"})

    page = await context.new_page()
    request = (
        await asyncio.gather(
            server.wait_for_request("/empty.html"),
            page.goto(server.EMPTY_PAGE),
        )
    )[0]
    await context.close()
    assert request.getHeader("foo") == "bar"


async def test_set_extra_http_headers_should_override_extra_headers_from_browser_context(
    browser: Browser, server: Server
) -> None:
    context = await browser.new_context(extra_http_headers={"fOo": "bAr", "baR": "foO"})

    page = await context.new_page()
    await page.set_extra_http_headers({"Foo": "Bar"})

    request = (
        await asyncio.gather(
            server.wait_for_request("/empty.html"),
            page.goto(server.EMPTY_PAGE),
        )
    )[0]
    await context.close()
    assert request.getHeader("foo") == "Bar"
    assert request.getHeader("bar") == "foO"


async def test_set_extra_http_headers_should_throw_for_non_string_header_values(
    page: Page,
) -> None:
    error: Optional[Error] = None
    try:
        await page.set_extra_http_headers({"foo": 1})  # type: ignore
    except Error as exc:
        error = exc
    assert error
    assert (
        error.message
        == "Page.set_extra_http_headers: headers[0].value: expected string, got number"
    )


async def test_response_server_addr(page: Page, server: Server) -> None:
    response = await page.goto(server.EMPTY_PAGE)
    assert response
    server_addr = await response.server_addr()
    assert server_addr
    assert server_addr["port"] == server.PORT
    assert server_addr["ipAddress"] in ["127.0.0.1", "[::1]"]


async def test_response_security_details(
    browser: Browser,
    https_server: Server,
    browser_name: str,
    is_win: bool,
    is_linux: bool,
) -> None:
    if (browser_name == "webkit" and is_linux) or (browser_name == "webkit" and is_win):
        pytest.skip("https://github.com/microsoft/playwright/issues/6759")
    page = await browser.new_page(ignore_https_errors=True)
    response = await page.goto(https_server.EMPTY_PAGE)
    assert response
    await response.finished()
    security_details = await response.security_details()
    assert security_details
    if browser_name == "webkit" and is_win:
        assert security_details == {
            "subjectName": "puppeteer-tests",
            "validFrom": 1550084863,
            "validTo": -1,
        }
    elif browser_name == "webkit":
        assert security_details == {
            "protocol": "TLS 1.3",
            "subjectName": "puppeteer-tests",
            "validFrom": 1550084863,
            "validTo": 33086084863,
        }
    else:
        assert security_details == {
            "issuer": "puppeteer-tests",
            "protocol": "TLS 1.3",
            "subjectName": "puppeteer-tests",
            "validFrom": 1550084863,
            "validTo": 33086084863,
        }
    await page.close()


async def test_response_security_details_none_without_https(
    page: Page, server: Server
) -> None:
    response = await page.goto(server.EMPTY_PAGE)
    assert response
    security_details = await response.security_details()
    assert security_details is None


async def test_should_report_if_request_was_from_service_worker(
    page: Page, server: Server
) -> None:
    response = await page.goto(server.PREFIX + "/serviceworkers/fetch/sw.html")
    assert response
    assert not response.from_service_worker
    await page.evaluate("() => window.activationPromise")
    async with page.expect_response("**/example.txt") as response_info:
        await page.evaluate("() => fetch('/example.txt')")
    response = await response_info.value
    assert response.from_service_worker



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio

from playwright.async_api import Page
from tests.server import Server

from .utils import Utils


async def test_should_think_that_it_is_focused_by_default(page: Page) -> None:
    assert await page.evaluate("document.hasFocus()")


async def test_should_think_that_all_pages_are_focused(page: Page) -> None:
    page2 = await page.context.new_page()
    assert await page.evaluate("document.hasFocus()")
    assert await page2.evaluate("document.hasFocus()")
    await page2.close()


async def test_should_focus_popups_by_default(page: Page, server: Server) -> None:
    await page.goto(server.EMPTY_PAGE)
    async with page.expect_popup() as popup_info:
        await page.evaluate("url => { window.open(url); }", server.EMPTY_PAGE)
    popup = await popup_info.value
    assert await popup.evaluate("document.hasFocus()")
    assert await page.evaluate("document.hasFocus()")


async def test_should_provide_target_for_keyboard_events(
    page: Page, server: Server
) -> None:
    page2 = await page.context.new_page()
    await asyncio.gather(
        page.goto(server.PREFIX + "/input/textarea.html"),
        page2.goto(server.PREFIX + "/input/textarea.html"),
    )
    await asyncio.gather(
        page.focus("input"),
        page2.focus("input"),
    )
    text = "first"
    text2 = "second"
    await asyncio.gather(
        page.keyboard.type(text),
        page2.keyboard.type(text2),
    )
    results = await asyncio.gather(
        page.evaluate("result"),
        page2.evaluate("result"),
    )
    assert results == [text, text2]


async def test_should_not_affect_mouse_event_target_page(
    page: Page, server: Server
) -> None:
    page2 = await page.context.new_page()
    click_counter = """() => {
      document.onclick = () => window.click_count = (window.click_count || 0) + 1;
    }"""
    await asyncio.gather(
        page.evaluate(click_counter),
        page2.evaluate(click_counter),
        page.focus("body"),
        page2.focus("body"),
    )
    await asyncio.gather(
        page.mouse.click(1, 1),
        page2.mouse.click(1, 1),
    )
    counters = await asyncio.gather(
        page.evaluate("window.click_count"),
        page2.evaluate("window.click_count"),
    )
    assert counters == [1, 1]


async def test_should_change_document_activeElement(page: Page, server: Server) -> None:
    page2 = await page.context.new_page()
    await asyncio.gather(
        page.goto(server.PREFIX + "/input/textarea.html"),
        page2.goto(server.PREFIX + "/input/textarea.html"),
    )
    await asyncio.gather(
        page.focus("input"),
        page2.focus("textarea"),
    )
    active = await asyncio.gather(
        page.evaluate("document.activeElement.tagName"),
        page2.evaluate("document.activeElement.tagName"),
    )
    assert active == ["INPUT", "TEXTAREA"]


async def test_should_change_focused_iframe(
    page: Page, server: Server, utils: Utils
) -> None:
    await page.goto(server.EMPTY_PAGE)
    [frame1, frame2] = await asyncio.gather(
        utils.attach_frame(page, "frame1", server.PREFIX + "/input/textarea.html"),
        utils.attach_frame(page, "frame2", server.PREFIX + "/input/textarea.html"),
    )
    logger = """() => {
        self._events = [];
        const element = document.querySelector('input');
        element.onfocus = element.onblur = (e) => self._events.push(e.type);
    }"""
    await asyncio.gather(
        frame1.evaluate(logger),
        frame2.evaluate(logger),
    )
    focused = await asyncio.gather(
        frame1.evaluate("document.hasFocus()"),
        frame2.evaluate("document.hasFocus()"),
    )
    assert focused == [False, False]
    await frame1.focus("input")
    events = await asyncio.gather(
        frame1.evaluate("self._events"),
        frame2.evaluate("self._events"),
    )
    assert events == [["focus"], []]
    focused = await asyncio.gather(
        frame1.evaluate("document.hasFocus()"),
        frame2.evaluate("document.hasFocus()"),
    )
    assert focused == [True, False]
    await frame2.focus("input")
    events = await asyncio.gather(
        frame1.evaluate("self._events"),
        frame2.evaluate("self._events"),
    )
    assert events == [["focus", "blur"], ["focus"]]
    focused = await asyncio.gather(
        frame1.evaluate("document.hasFocus()"),
        frame2.evaluate("document.hasFocus()"),
    )
    assert focused == [False, True]



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Generator

import pytest

from playwright.sync_api import Page, expect

from .utils import (
    TODO_ITEMS,
    check_number_of_completed_todos_in_local_storage,
    check_todos_in_local_storage,
    create_default_todos,
)


@pytest.fixture(autouse=True)
def run_around_tests(page: Page) -> Generator[None, None, None]:
    # setup before a test
    page.goto("https://demo.playwright.dev/todomvc")
    create_default_todos(page)
    # make sure the app had a chance to save updated todos in storage
    # before navigating to a new view, otherwise the items can get lost :(
    # in some frameworks like Durandal
    check_todos_in_local_storage(page, TODO_ITEMS[0])
    # run the actual test
    yield
    # run any cleanup code


def test_should_allow_me_to_display_active_item(page: Page) -> None:
    page.locator(".todo-list li .toggle").nth(1).check()
    check_number_of_completed_todos_in_local_storage(page, 1)
    page.locator(".filters >> text=Active").click()
    expect(page.locator(".todo-list li")).to_have_count(2)
    expect(page.locator(".todo-list li")).to_have_text([TODO_ITEMS[0], TODO_ITEMS[2]])


def test_should_respect_the_back_button(page: Page) -> None:
    page.locator(".todo-list li .toggle").nth(1).check()
    check_number_of_completed_todos_in_local_storage(page, 1)

    # Showing all items
    page.locator(".filters >> text=All").click()
    expect(page.locator(".todo-list li")).to_have_count(3)

    # Showing active items
    page.locator(".filters >> text=Active").click()

    # Showing completed items
    page.locator(".filters >> text=Completed").click()

    expect(page.locator(".todo-list li")).to_have_count(1)
    page.go_back()
    expect(page.locator(".todo-list li")).to_have_count(2)
    page.go_back()
    expect(page.locator(".todo-list li")).to_have_count(3)


def test_should_allow_me_to_display_completed_items(page: Page) -> None:
    page.locator(".todo-list li .toggle").nth(1).check()
    check_number_of_completed_todos_in_local_storage(page, 1)
    page.locator(".filters >> text=Completed").click()
    expect(page.locator(".todo-list li")).to_have_count(1)


def test_should_allow_me_to_display_all_items(page: Page) -> None:
    page.locator(".todo-list li .toggle").nth(1).check()
    check_number_of_completed_todos_in_local_storage(page, 1)
    page.locator(".filters >> text=Active").click()
    page.locator(".filters >> text=Completed").click()
    page.locator(".filters >> text=All").click()
    expect(page.locator(".todo-list li")).to_have_count(3)


def test_should_highlight_the_current_applied_filter(page: Page) -> None:
    expect(page.locator(".filters >> text=All")).to_have_class("selected")
    page.locator(".filters >> text=Active").click()
    # Page change - active items.
    expect(page.locator(".filters >> text=Active")).to_have_class("selected")
    page.locator(".filters >> text=Completed").click()
    # Page change - completed items.
    expect(page.locator(".filters >> text=Completed")).to_have_class("selected")



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

import pytest

from playwright.sync_api import BrowserType
from tests.server import find_free_port

pytestmark = pytest.mark.only_browser("chromium")


def test_connect_to_an_existing_cdp_session(
    launch_arguments: Dict, browser_type: BrowserType
) -> None:
    port = find_free_port()
    browser_server = browser_type.launch(
        **launch_arguments, args=[f"--remote-debugging-port={port}"]
    )
    cdp_browser = browser_type.connect_over_cdp(f"http://127.0.0.1:{port}")
    assert len(cdp_browser.contexts) == 1
    cdp_browser.close()
    browser_server.close()



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, TypeVar

TARGET_CLOSED_ERROR_MESSAGE = "Target page, context or browser has been closed"

MustType = TypeVar("MustType")


def must(value: Optional[MustType]) -> MustType:
    assert value
    return value


def chromium_version_less_than(a: str, b: str) -> bool:
    left = list(map(int, a.split(".")))
    right = list(map(int, b.split(".")))
    for i in range(4):
        if left[i] > right[i]:
            return False
        if left[i] < right[i]:
            return True
    return False



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Generator

import pytest

from playwright.sync_api import Page, expect

from .utils import (
    TODO_ITEMS,
    assert_number_of_todos_in_local_storage,
    check_todos_in_local_storage,
    create_default_todos,
)


@pytest.fixture(autouse=True)
def run_around_tests(page: Page) -> Generator[None, None, None]:
    # setup before a test
    page.goto("https://demo.playwright.dev/todomvc")
    create_default_todos(page)
    assert_number_of_todos_in_local_storage(page, 3)
    # run the actual test
    yield
    # run any cleanup code


def test_should_hide_other_controls_when_editing(page: Page) -> None:
    todo_item = page.locator(".todo-list li").nth(1)
    todo_item.dblclick()
    expect(todo_item.locator(".toggle")).not_to_be_visible()
    expect(todo_item.locator("label")).not_to_be_visible()
    assert_number_of_todos_in_local_storage(page, 3)


def test_should_save_edits_on_blur(page: Page) -> None:
    todo_items = page.locator(".todo-list li")
    todo_items.nth(1).dblclick()
    todo_items.nth(1).locator(".edit").fill("buy some sausages")
    todo_items.nth(1).locator(".edit").dispatch_event("blur")

    expect(todo_items).to_have_text(
        [
            TODO_ITEMS[0],
            "buy some sausages",
            TODO_ITEMS[2],
        ]
    )
    check_todos_in_local_storage(page, "buy some sausages")


def test_should_trim_entered_text(page: Page) -> None:
    todo_items = page.locator(".todo-list li")
    todo_items.nth(1).dblclick()
    todo_items.nth(1).locator(".edit").fill("    buy some sausages    ")
    todo_items.nth(1).locator(".edit").press("Enter")

    expect(todo_items).to_have_text(
        [
            TODO_ITEMS[0],
            "buy some sausages",
            TODO_ITEMS[2],
        ]
    )
    check_todos_in_local_storage(page, "buy some sausages")


def test_should_remove_the_item_if_an_empty_text_string_was_entered(page: Page) -> None:
    todo_items = page.locator(".todo-list li")
    todo_items.nth(1).dblclick()
    todo_items.nth(1).locator(".edit").fill("")
    todo_items.nth(1).locator(".edit").press("Enter")

    expect(todo_items).to_have_text(
        [
            TODO_ITEMS[0],
            TODO_ITEMS[2],
        ]
    )


def test_should_cancel_edits_on_escape(page: Page) -> None:
    todo_items = page.locator(".todo-list li")
    todo_items.nth(1).dblclick()
    todo_items.nth(1).locator(".edit").press("Escape")
    expect(todo_items).to_have_text(TODO_ITEMS)



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from playwright.async_api import Error, Page, Playwright


async def test_should_use_data_test_id_in_strict_errors(
    page: Page, playwright: Playwright
) -> None:
    playwright.selectors.set_test_id_attribute("data-custom-id")
    try:
        await page.set_content(
            """
          <div>
            <div></div>
            <div>
              <div></div>
              <div></div>
            </div>
          </div>
          <div>
            <div class='foo bar:0' data-custom-id='One'>
            </div>
            <div class='foo bar:1' data-custom-id='Two'>
            </div>
          </div>
        """
        )
        with pytest.raises(Error) as exc_info:
            await page.locator(".foo").hover()
        assert "strict mode violation" in exc_info.value.message
        assert '<div class="foo bar:0' in exc_info.value.message
        assert '<div class="foo bar:1' in exc_info.value.message
        assert 'aka get_by_test_id("One")' in exc_info.value.message
        assert 'aka get_by_test_id("Two")' in exc_info.value.message
    finally:
        playwright.selectors.set_test_id_attribute("data-testid")



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from playwright.async_api import Page, Selectors
from tests.server import Server

from .utils import Utils


async def test_should_dispatch_click_event(page: Page, server: Server) -> None:
    await page.goto(server.PREFIX + "/input/button.html")
    await page.dispatch_event("button", "click")
    assert await page.evaluate("() => result") == "Clicked"


async def test_should_dispatch_click_event_properties(
    page: Page, server: Server
) -> None:
    await page.goto(server.PREFIX + "/input/button.html")
    await page.dispatch_event("button", "click")
    assert await page.evaluate("() => bubbles")
    assert await page.evaluate("() => cancelable")
    assert await page.evaluate("() => composed")


async def test_should_dispatch_click_svg(page: Page) -> None:
    await page.set_content(
        """
      <svg height="100" width="100">
        <circle onclick="javascript:window.__CLICKED=42" cx="50" cy="50" r="40" stroke="black" stroke-width="3" fill="red" />
      </svg>
    """
    )
    await page.dispatch_event("circle", "click")
    assert await page.evaluate("() => window.__CLICKED") == 42


async def test_should_dispatch_click_on_a_span_with_an_inline_element_inside(
    page: Page,
) -> None:
    await page.set_content(
        """
      <style>
      span::before {
        content: 'q';
      }
      </style>
      <span onclick='javascript:window.CLICKED=42'></span>
    """
    )
    await page.dispatch_event("span", "click")
    assert await page.evaluate("() => window.CLICKED") == 42


async def test_should_dispatch_click_after_navigation(
    page: Page, server: Server
) -> None:
    await page.goto(server.PREFIX + "/input/button.html")
    await page.dispatch_event("button", "click")
    await page.goto(server.PREFIX + "/input/button.html")
    await page.dispatch_event("button", "click")
    assert await page.evaluate("() => result") == "Clicked"


async def test_should_dispatch_click_after_a_cross_origin_navigation(
    page: Page, server: Server
) -> None:
    await page.goto(server.PREFIX + "/input/button.html")
    await page.dispatch_event("button", "click")
    await page.goto(server.CROSS_PROCESS_PREFIX + "/input/button.html")
    await page.dispatch_event("button", "click")
    assert await page.evaluate("() => result") == "Clicked"


async def test_should_not_fail_when_element_is_blocked_on_hover(page: Page) -> None:
    await page.set_content(
        """<style>
      container { display: block; position: relative; width: 200px; height: 50px; }
      div, button { position: absolute; left: 0; top: 0; bottom: 0; right: 0; }
      div { pointer-events: none; }
      container:hover div { pointer-events: auto; background: red; }
    </style>
    <container>
      <button onclick="window.clicked=true">Click me</button>
      <div></div>
    </container>"""
    )
    await page.dispatch_event("button", "click")
    assert await page.evaluate("() => window.clicked")


async def test_should_dispatch_click_when_node_is_added_in_shadow_dom(
    page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    watchdog = page.dispatch_event("span", "click")
    await page.evaluate(
        """() => {
      const div = document.createElement('div');
      div.attachShadow({mode: 'open'});
      document.body.appendChild(div);
    }"""
    )
    await page.evaluate("() => new Promise(f => setTimeout(f, 100))")
    await page.evaluate(
        """() => {
      const span = document.createElement('span');
      span.textContent = 'Hello from shadow';
      span.addEventListener('click', () => window.clicked = true);
      document.querySelector('div').shadowRoot.appendChild(span);
    }"""
    )
    await watchdog
    assert await page.evaluate("() => window.clicked")


async def test_should_be_atomic(selectors: Selectors, page: Page, utils: Utils) -> None:
    await utils.register_selector_engine(
        selectors,
        "dispatch_event",
        """{
            create(root, target) { },
            query(root, selector) {
                const result = root.querySelector(selector);
                if (result)
                Promise.resolve().then(() => result.onclick = "");
                return result;
            },
            queryAll(root, selector) {
                const result = Array.from(root.querySelectorAll(selector));
                for (const e of result)
                Promise.resolve().then(() => result.onclick = "");
                return result;
            }
        }""",
    )
    await page.set_content('<div onclick="window._clicked=true">Hello</div>')
    await page.dispatch_event("dispatch_event=div", "click")
    assert await page.evaluate("() => window._clicked")


async def test_should_dispatch_drag_drop_events(page: Page, server: Server) -> None:
    await page.goto(server.PREFIX + "/drag-n-drop.html")
    dataTransfer = await page.evaluate_handle("() => new DataTransfer()")
    await page.dispatch_event("#source", "dragstart", {"dataTransfer": dataTransfer})
    await page.dispatch_event("#target", "drop", {"dataTransfer": dataTransfer})
    assert await page.evaluate(
        """() => {
      return source.parentElement === target;
    }"""
    )


async def test_should_dispatch_drag_and_drop_events_element_handle(
    page: Page, server: Server
) -> None:
    await page.goto(server.PREFIX + "/drag-n-drop.html")
    dataTransfer = await page.evaluate_handle("() => new DataTransfer()")
    source = await page.query_selector("#source")
    assert source
    await source.dispatch_event("dragstart", {"dataTransfer": dataTransfer})
    target = await page.query_selector("#target")
    assert target
    await target.dispatch_event("drop", {"dataTransfer": dataTransfer})
    assert await page.evaluate(
        """() => {
      return source.parentElement === target;
    }"""
    )


async def test_should_dispatch_click_event_element_handle(
    page: Page, server: Server
) -> None:
    await page.goto(server.PREFIX + "/input/button.html")
    button = await page.query_selector("button")
    assert button
    await button.dispatch_event("click")
    assert await page.evaluate("() => result") == "Clicked"



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio

import pytest

from playwright.async_api import Error, Page, Request
from tests.server import Server


async def test_should_not_allow_to_access_frame_on_popup_main_request(
    page: Page, server: Server
) -> None:
    await page.set_content(f'<a target=_blank href="{server.EMPTY_PAGE}">click me</a>')
    request_promise = asyncio.ensure_future(page.context.wait_for_event("request"))
    popup_promise = asyncio.ensure_future(page.context.wait_for_event("page"))
    clicked = asyncio.ensure_future(page.get_by_text("click me").click())
    request: Request = await request_promise

    assert request.is_navigation_request()

    with pytest.raises(Error) as exc_info:
        request.frame
    assert (
        "Frame for this navigation request is not available" in exc_info.value.message
    )

    response = await request.response()
    assert response
    await response.finished()
    await popup_promise
    await clicked


async def test_should_parse_the_data_if_content_type_is_application_x_www_form_urlencoded_charset_UTF_8(
    page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    async with page.expect_event("request") as request_info:
        await page.evaluate(
            """() => fetch('./post', {
            method: 'POST',
            headers: {
            'content-type': 'application/x-www-form-urlencoded; charset=UTF-8',
            },
            body: 'foo=bar&baz=123'
        })"""
        )
    request = await request_info.value
    assert request
    assert request.post_data_json == {"foo": "bar", "baz": "123"}



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Any, Callable, Dict, Generator, List, Optional

import pytest

from playwright.sync_api import BrowserContext, BrowserType


@pytest.fixture()
def launch_persistent_context(
    browser_type: BrowserType,
    browser_channel: Optional[str],
    tmp_path: Path,
    launch_arguments: Dict[str, Any],
    is_headless_shell: bool,
) -> Generator[Callable[..., BrowserContext], None, None]:
    if browser_channel and browser_channel.startswith("chrome"):
        pytest.skip(
            "--load-extension is not supported in Chrome anymore. https://groups.google.com/a/chromium.org/g/chromium-extensions/c/1-g8EFx2BBY/m/S0ET5wPjCAAJ"
        )
    if is_headless_shell:
        pytest.skip("Headless Shell has no support for extensions")

    contexts: List[BrowserContext] = []

    def launch(extension_path: str, **kwargs: Any) -> BrowserContext:
        context = browser_type.launch_persistent_context(
            str(tmp_path),
            **launch_arguments,
            **kwargs,
            args=[
                f"--disable-extensions-except={extension_path}",
                f"--load-extension={extension_path}",
            ],
        )
        contexts.append(context)
        return context

    yield launch

    for context in contexts:
        context.close()


@pytest.mark.only_browser("chromium")
def test_should_give_access_to_the_service_worker(
    launch_persistent_context: Callable[..., BrowserContext],
    assetdir: Path,
) -> None:
    extension_path = str(assetdir / "extension-mv3-simple")
    context = launch_persistent_context(extension_path)
    service_workers = context.service_workers
    service_worker = (
        service_workers[0]
        if len(service_workers)
        else context.wait_for_event("serviceworker")
    )
    assert service_worker
    assert service_worker in context.service_workers
    while not service_worker.evaluate("globalThis.MAGIC") == 42:
        context.pages[0].wait_for_timeout(100)
    context.close()
    assert len(context.background_pages) == 0


@pytest.mark.only_browser("chromium")
def test_should_give_access_to_the_service_worker_when_recording_video(
    launch_persistent_context: Callable[..., BrowserContext],
    tmp_path: Path,
    assetdir: Path,
) -> None:
    extension_path = str(assetdir / "extension-mv3-simple")
    context = launch_persistent_context(
        extension_path, record_video_dir=(tmp_path / "videos")
    )
    service_workers = context.service_workers
    service_worker = (
        service_workers[0]
        if len(service_workers)
        else context.wait_for_event("serviceworker")
    )
    assert service_worker
    assert service_worker in context.service_workers
    while not service_worker.evaluate("globalThis.MAGIC") == 42:
        context.pages[0].wait_for_timeout(100)
    context.close()
    assert len(context.background_pages) == 0



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
from typing import Dict

from playwright.sync_api import sync_playwright


def test_running_in_thread(browser_name: str, launch_arguments: Dict) -> None:
    result = []

    class TestThread(threading.Thread):
        def run(self) -> None:
            with sync_playwright() as playwright:
                browser = playwright[browser_name].launch(**launch_arguments)
                # This should not throw ^^.
                browser.new_page()
                browser.close()
                result.append("Success")

    test_thread = TestThread()
    test_thread.start()
    test_thread.join()
    assert "Success" in result



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import io
import json
import os
import subprocess
import sys
from pathlib import Path
from typing import Any, Callable, Dict, Generator, List, Optional, cast

import pytest
from PIL import Image
from pixelmatch import pixelmatch
from pixelmatch.contrib.PIL import from_PIL_to_raw_data

import playwright
from playwright._impl._path_utils import get_file_dirname

from .server import Server, test_server

_dirname = get_file_dirname()


def pytest_configure(config: pytest.Config) -> None:
    if os.environ.get("CI"):
        config.option.reruns = 3


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    if "browser_name" in metafunc.fixturenames:
        browsers = metafunc.config.option.browser or ["chromium", "firefox", "webkit"]
        metafunc.parametrize("browser_name", browsers, scope="session")


@pytest.fixture(scope="session")
def assetdir() -> Path:
    return _dirname / "assets"


@pytest.fixture(scope="session")
def headless(pytestconfig: pytest.Config) -> bool:
    return not (pytestconfig.getoption("--headed") or os.getenv("HEADFUL", False))


@pytest.fixture(scope="session")
def launch_arguments(pytestconfig: pytest.Config, headless: bool) -> Dict:
    args: Dict = {
        "headless": headless,
    }
    if pytestconfig.getoption("--browser-channel"):
        args["channel"] = pytestconfig.getoption("--browser-channel")
    return args


@pytest.fixture
def server() -> Generator[Server, None, None]:
    yield test_server.server


@pytest.fixture
def https_server() -> Generator[Server, None, None]:
    yield test_server.https_server


@pytest.fixture(autouse=True, scope="session")
def start_server() -> Generator[None, None, None]:
    test_server.start()
    yield
    test_server.stop()


@pytest.fixture(autouse=True)
def after_each_hook() -> Generator[None, None, None]:
    yield
    test_server.reset()


@pytest.fixture(scope="session")
def browser_name(pytestconfig: pytest.Config) -> str:
    return cast(str, pytestconfig.getoption("browser"))


@pytest.fixture(scope="session")
def browser_channel(pytestconfig: pytest.Config) -> Optional[str]:
    return cast(Optional[str], pytestconfig.getoption("--browser-channel"))


@pytest.fixture(scope="session")
def is_headless_shell(browser_name: str, browser_channel: str, headless: bool) -> bool:
    return browser_name == "chromium" and (
        browser_channel == "chromium-headless-shell"
        or (not browser_channel and headless)
    )


@pytest.fixture(scope="session")
def is_webkit(browser_name: str) -> bool:
    return browser_name == "webkit"


@pytest.fixture(scope="session")
def is_firefox(browser_name: str) -> bool:
    return browser_name == "firefox"


@pytest.fixture(scope="session")
def is_chromium(browser_name: str) -> bool:
    return browser_name == "chromium"


@pytest.fixture(scope="session")
def is_win() -> bool:
    return sys.platform == "win32"


@pytest.fixture(scope="session")
def is_linux() -> bool:
    return sys.platform == "linux"


@pytest.fixture(scope="session")
def is_mac() -> bool:
    return sys.platform == "darwin"


def _get_skiplist(
    request: pytest.FixtureRequest, values: List[str], value_name: str
) -> List[str]:
    skipped_values = []
    # Allowlist
    only_marker = request.node.get_closest_marker(f"only_{value_name}")
    if only_marker:
        skipped_values = values
        skipped_values.remove(only_marker.args[0])

    # Denylist
    skip_marker = request.node.get_closest_marker(f"skip_{value_name}")
    if skip_marker:
        skipped_values.append(skip_marker.args[0])

    return skipped_values


@pytest.fixture(autouse=True)
def skip_by_browser(request: pytest.FixtureRequest, browser_name: str) -> None:
    skip_browsers_names = _get_skiplist(
        request, ["chromium", "firefox", "webkit"], "browser"
    )

    if browser_name in skip_browsers_names:
        pytest.skip(f"skipped for this browser: {browser_name}")


@pytest.fixture(autouse=True)
def skip_by_platform(request: pytest.FixtureRequest) -> None:
    skip_platform_names = _get_skiplist(
        request, ["win32", "linux", "darwin"], "platform"
    )

    if sys.platform in skip_platform_names:
        pytest.skip(f"skipped on this platform: {sys.platform}")


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup("playwright", "Playwright")
    group.addoption(
        "--browser",
        action="append",
        default=[],
        help="Browsers which should be used. By default on all the browsers.",
    )
    group.addoption(
        "--browser-channel",
        action="store",
        default=None,
        help="Browser channel to be used.",
    )
    parser.addoption(
        "--headed",
        action="store_true",
        default=False,
        help="Run tests in headed mode.",
    )


@pytest.fixture(scope="session")
def assert_to_be_golden(browser_name: str) -> Callable[[bytes, str], None]:
    def compare(received_raw: bytes, golden_name: str) -> None:
        golden_file_path = _dirname / f"golden-{browser_name}" / golden_name
        try:
            golden_file = golden_file_path.read_bytes()
            received_image = Image.open(io.BytesIO(received_raw))
            golden_image = Image.open(io.BytesIO(golden_file))

            if golden_image.size != received_image.size:
                pytest.fail("Image size differs to golden image")
                return
            diff_pixels = pixelmatch(
                from_PIL_to_raw_data(received_image),
                from_PIL_to_raw_data(golden_image),
                golden_image.size[0],
                golden_image.size[1],
                threshold=0.2,
            )
            assert diff_pixels == 0
        except Exception:
            if os.getenv("PW_WRITE_SCREENSHOT"):
                golden_file_path.parent.mkdir(parents=True, exist_ok=True)
                golden_file_path.write_bytes(received_raw)
                print(f"Wrote {golden_file_path}")
            raise

    return compare


class RemoteServer:
    def __init__(
        self, browser_name: str, launch_server_options: Dict, tmpfile: Path
    ) -> None:
        driver_dir = Path(inspect.getfile(playwright)).parent / "driver"
        if sys.platform == "win32":
            node_executable = driver_dir / "node.exe"
        else:
            node_executable = driver_dir / "node"
        cli_js = driver_dir / "package" / "cli.js"
        tmpfile.write_text(json.dumps(launch_server_options))
        self.process = subprocess.Popen(
            [
                str(node_executable),
                str(cli_js),
                "launch-server",
                "--browser",
                browser_name,
                "--config",
                str(tmpfile),
            ],
            stdout=subprocess.PIPE,
            stderr=sys.stderr,
            cwd=driver_dir,
        )
        assert self.process.stdout
        self.ws_endpoint = self.process.stdout.readline().decode().strip()
        self.process.stdout.close()

    def kill(self) -> None:
        # Send the signal to all the process groups
        if self.process.poll() is not None:
            return
        self.process.kill()
        self.process.wait()


@pytest.fixture
def launch_server(
    browser_name: str, launch_arguments: Dict, tmp_path: Path
) -> Generator[Callable[..., RemoteServer], None, None]:
    remotes: List[RemoteServer] = []

    def _launch_server(**kwargs: Dict[str, Any]) -> RemoteServer:
        remote = RemoteServer(
            browser_name,
            {
                **launch_arguments,
                **kwargs,
            },
            tmp_path / f"settings-{len(remotes)}.json",
        )
        remotes.append(remote)
        return remote

    yield _launch_server

    for remote in remotes:
        remote.kill()



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from typing import Dict

import pytest
import requests

from playwright.async_api import BrowserType, Error
from tests.server import Server, WebSocketProtocol, find_free_port

pytestmark = pytest.mark.only_browser("chromium")


async def test_connect_to_an_existing_cdp_session(
    launch_arguments: Dict, browser_type: BrowserType
) -> None:
    port = find_free_port()
    browser_server = await browser_type.launch(
        **launch_arguments, args=[f"--remote-debugging-port={port}"]
    )
    cdp_browser = await browser_type.connect_over_cdp(f"http://127.0.0.1:{port}")
    assert len(cdp_browser.contexts) == 1
    await cdp_browser.close()
    await browser_server.close()


async def test_connect_to_an_existing_cdp_session_twice(
    launch_arguments: Dict, browser_type: BrowserType, server: Server
) -> None:
    port = find_free_port()
    browser_server = await browser_type.launch(
        **launch_arguments, args=[f"--remote-debugging-port={port}"]
    )
    endpoint_url = f"http://127.0.0.1:{port}"
    cdp_browser1 = await browser_type.connect_over_cdp(endpoint_url)
    cdp_browser2 = await browser_type.connect_over_cdp(endpoint_url)
    assert len(cdp_browser1.contexts) == 1
    page1 = await cdp_browser1.contexts[0].new_page()
    await page1.goto(server.EMPTY_PAGE)

    assert len(cdp_browser2.contexts) == 1
    page2 = await cdp_browser2.contexts[0].new_page()
    await page2.goto(server.EMPTY_PAGE)

    assert len(cdp_browser1.contexts[0].pages) == 2
    assert len(cdp_browser2.contexts[0].pages) == 2

    await cdp_browser1.close()
    await cdp_browser2.close()
    await browser_server.close()


def _ws_endpoint_from_url(url: str) -> str:
    response = requests.get(url)
    assert response.ok
    response_body = response.json()
    return response_body["webSocketDebuggerUrl"]


async def test_conect_over_a_ws_endpoint(
    launch_arguments: Dict, browser_type: BrowserType, server: Server
) -> None:
    port = find_free_port()
    browser_server = await browser_type.launch(
        **launch_arguments, args=[f"--remote-debugging-port={port}"]
    )
    ws_endpoint = _ws_endpoint_from_url(f"http://127.0.0.1:{port}/json/version/")

    cdp_browser1 = await browser_type.connect_over_cdp(ws_endpoint)
    assert len(cdp_browser1.contexts) == 1
    await cdp_browser1.close()

    cdp_browser2 = await browser_type.connect_over_cdp(ws_endpoint)
    assert len(cdp_browser2.contexts) == 1
    await cdp_browser2.close()
    await browser_server.close()


async def test_connect_over_cdp_passing_header_works(
    browser_type: BrowserType, server: Server
) -> None:
    server.send_on_web_socket_connection(b"incoming")
    request = asyncio.create_task(server.wait_for_request("/ws"))
    with pytest.raises(Error):
        await browser_type.connect_over_cdp(
            f"ws://127.0.0.1:{server.PORT}/ws", headers={"foo": "bar"}
        )
    assert (await request).getHeader("foo") == "bar"


async def test_should_print_custom_ws_close_error(
    browser_type: BrowserType, server: Server
) -> None:
    def _handle_ws(ws: WebSocketProtocol) -> None:
        def _onMessage(payload: bytes, isBinary: bool) -> None:
            ws.sendClose(code=4123, reason="Oh my!")

        setattr(ws, "onMessage", _onMessage)

    server.once_web_socket_connection(_handle_ws)
    with pytest.raises(Error, match="Browser logs:\n\nOh my!\n"):
        await browser_type.connect_over_cdp(
            f"ws://127.0.0.1:{server.PORT}/ws", headers={"foo": "bar"}
        )
